import os
import ray
import time
import torch
import queue
import threading
from shutil import copy2
from copy import deepcopy

import pytorchrl as prl
from pytorchrl.scheme.base.worker import Worker as W
from pytorchrl.scheme.utils import ray_get_and_free, broadcast_message, pack, unpack

# Puts a limit to the allowed policy lag
max_queue_size = 10


class GWorker(W):
    """
    Worker class handling gradient computation.

    This class wraps an actor instance, a storage class instance and a
    worker set of remote data collection workers. It receives data from the
    collection workers and computes gradients following a logic defined in
    function self.step(), which will be called from the Learner class.

    Parameters
    ----------
    index_worker : int
        Worker index.
    col_workers_factory : func
        A function that creates a set of data collection workers.
    col_communication : str
        Communication coordination pattern for data collection.
    compress_grads_to_send : bool
        Whether or not to compress gradients before sending then to the update worker.
    col_execution : str
        Execution patterns for data collection.
    col_fraction_workers : float
        Minimum fraction of samples required to stop if collection is
        synchronously coordinated and most workers have finished their
        collection task.
    device : str
        "cpu" or specific GPU "cuda:number`" to use for computation.
    initial_weights : ray object ID
        Initial model weights.

    Attributes
    ----------
    index_worker : int
        Index assigned to this worker.
    iter : int
        Number of times gradients have been computed and sent.
    col_communication : str
        Communication coordination pattern for data collection.
    compress_grads_to_send : bool
        Whether or not to compress gradients before sending then to the update worker.
    col_workers : CWorkerSet
        A CWorkerSet class instance.
    local_worker : CWorker
        col_workers local worker.
    remote_workers : List of CWorker's
        col_workers remote data collection workers.
    actor : Actor
        An actor class instance.
    algo : Algo
        An algorithm class instance.
    storage : Storage
        A Storage class instance.
    inqueue : queue.Queue
        Input queue where incoming collected samples are placed.
    collector : CollectorThread
        Class handling data collection via col_workers and placing incoming
        rollouts into the input queue `inqueue`.
    """

    def __init__(self,
                 index_worker,
                 col_workers_factory,
                 col_communication=prl.SYNC,
                 compress_grads_to_send=False,
                 col_execution=prl.CENTRAL,
                 col_fraction_workers=1.0,
                 initial_weights=None,
                 device=None):

        self.index_worker = index_worker
        super(GWorker, self).__init__(index_worker)

        # Define counters and other attributes
        self.iter = 0
        self.col_communication = col_communication
        self.compress_grads_to_send = compress_grads_to_send
        self.processing_time_start = None

        # Computation device
        dev = device or "cuda" if torch.cuda.is_available() else "cpu"

        # Create CWorkerSet instance
        self.col_workers = col_workers_factory(dev, initial_weights, index_worker)
        self.local_worker = self.col_workers.local_worker()
        self.remote_workers = self.col_workers.remote_workers()
        self.num_remote_workers = len(self.remote_workers)
        self.num_collection_workers = 1 if self.num_remote_workers == 0 else self.num_remote_workers

        # Get Actor Critic instance
        self.actor = self.local_worker.actor

        # Get Algorithm instance
        self.algo = self.local_worker.algo

        # Get storage instance.
        if col_communication == prl.ASYNC and self.local_worker.envs_train is not None:

            # If async collection, c_worker and g_worker need different storage instances
            # To avoid overwriting data. Define c_worker one with the minimum size required,
            # to save memory
            size1 = self.local_worker.storage.max_size
            size2 = self.local_worker.algo.update_every
            size2 = size2 * 2 if size2 is not None else float("Inf")
            new_size = min(size1, size2)

            if self.local_worker.envs_train is not None:

                # Create a storage class deepcopy, but be careful with envs attribute,
                # can not be deep-copied
                envs = getattr(self.local_worker.storage, "envs")
                self.local_worker.storage.envs = None
                self.storage = deepcopy(self.local_worker.storage)
                self.local_worker.storage.envs = envs
                self.local_worker.update_storage_parameter("max_size", new_size)
            else:
                self.storage = self.local_worker.storage
                for e in self.remote_workers:
                    e.update_storage_parameter.remote("max_size", new_size)
        else:
            self.storage = self.local_worker.storage

        if len(self.remote_workers) > 0 or (len(self.remote_workers) == 0 and
                self.local_worker.envs_train is not None):

            # Create CollectorThread
            self.collector = CollectorThread(
                index_worker=index_worker,
                local_worker=self.local_worker,
                remote_workers=self.remote_workers,
                col_communication=col_communication,
                col_fraction_workers=col_fraction_workers,
                col_execution=col_execution,
                broadcast_interval=1)

            # Print worker information
            self.print_worker_info()

    @property
    def actor_version(self):
        """Number of times Actor has been updated."""
        return self.local_worker.actor_version

    def step(self, distribute_gradients=False):
        """
        Pulls data from `self.collector.queue`, then perform a gradient computation step.

        Parameters
        ----------
        distribute_gradients : bool
            If True, gradients will be directly shared across remote workers
            and optimization steps will executed in a decentralised way.

        Returns
        -------
        grads : list of tensors
            List of actor gradients.
        info : dict
            Summary dict of relevant gradient operation information.
        """

        self.get_data()
        grads, info = self.get_grads(distribute_gradients)

        if distribute_gradients:
            self.apply_gradients()

        # Encode data if self.compress_data_to_send is True
        data_to_send = pack((grads, info)) if self.compress_grads_to_send else (grads, info)

        return data_to_send

    def get_data(self):
        """
        Pulls data from `self.collector.queue` and prepares batches to
        compute gradients.
        """

        try:  # Try pulling next batch
            self.next_batch = self.batches.__next__()
            # Only use episode information once
            self.info.pop(prl.EPISODES, None)
            # Only record collected samples once
            self.info[prl.NUMSAMPLES] = 0

        # except (StopIteration, AttributeError):
        except Exception:

            # Check if more batches with the same data are required
            can_reuse_data = hasattr(self.algo, "reuse_data")
            if not can_reuse_data or not self.algo.reuse_data:

                # Get new data
                if self.col_communication == prl.SYNC:
                    self.collector.step()
                data, self.info = self.collector.queue.get()

                # Calculate and log data collection-to-processing time ratio
                if not self.processing_time_start:
                    self.processing_time_start = time.time()
                else:
                    self.info[prl.TIME][prl.PROCESSING] = time.time() - self.processing_time_start
                    self.info[prl.TIME][prl.CPRATIO] = (
                        self.info[prl.TIME][prl.COLLECTION] / self.num_collection_workers) / self.info[prl.TIME][
                        prl.PROCESSING]
                    self.processing_time_start = time.time()

                # Proprocess new data
                self.storage.insert_data_slice(data)
                self.storage.before_gradients()
            else:
                # Only record collected samples once
                self.info[prl.NUMSAMPLES] = 0

            # Genarate batches
            self.batches = self.storage.generate_batches(
                self.algo.num_mini_batch,
                self.algo.mini_batch_size,
                self.algo.num_epochs)
            self.next_batch = self.batches.__next__()

    def get_grads(self, distribute_gradients=False):
        """
        Perform a gradient computation step.

        Parameters
        ----------
        distribute_gradients : bool
            If True, gradients will be directly shared across remote workers
            and optimization steps will executed in a decentralised way.

        Returns
        -------
        grads : list of tensors
            List of actor gradients.
        info : dict
            Summary dict of relevant gradient operation information.
        """

        # Get gradients and algorithm-related information
        t = time.time()
        grads, algo_info = self.compute_gradients(self.next_batch, distribute_gradients)
        compute_time = time.time() - t

        # Add extra information to info dict
        self.info[prl.ALGORITHM] = algo_info
        self.info[prl.VERSION][prl.GRADIENT] = self.local_worker.actor_version
        self.info[prl.TIME][prl.GRADIENT] = compute_time

        # Run after gradients data process (if any)
        info = self.storage.after_gradients(self.next_batch, self.info)

        # Update iteration counter
        self.iter += 1

        return grads, info

    def compute_gradients(self, batch, distribute_gradients):
        """
        Calculate actor gradients and update networks.

        Parameters
        ----------
        batch : dict
            data batch containing all required tensors to compute algo loss.
        distribute_gradients : bool
            If True, gradients will be directly shared across remote workers
            and optimization steps will executed in a decentralised way.

        Returns
        -------
        grads : list of tensors
            List of actor gradients.
        info : dict
            Summary dict with relevant gradient-related information.
        """

        grads, info = self.algo.compute_gradients(batch, grads_to_cpu=not distribute_gradients)

        if distribute_gradients:

            if torch.cuda.is_available():
                for g in grads:
                    torch.distributed.all_reduce(g, op=torch.distributed.ReduceOp.SUM)
            else:
                torch.distributed.all_reduce_coalesced(grads, op=torch.distributed.ReduceOp.SUM)

            for p in self.actor.parameters():
                if p.grad is not None:
                    p.grad /= self.num_remote_workers

            grads = None

        return grads, info

    def apply_gradients(self, gradients=None):
        """Update Actor Critic model"""

        self.local_worker.actor_version += 1
        self.algo.apply_gradients(gradients)
        if self.col_communication == prl.SYNC and len(self.remote_workers) > 0:
            self.collector.broadcast_new_weights()

    def set_weights(self, actor_weights):
        """
        Update the worker actor version with provided weights.

        weights : dict of tensors
            Dict containing actor weights to be set.
        """

        self.local_worker.actor_version = actor_weights[prl.VERSION]
        self.local_worker.algo.set_weights(actor_weights[prl.WEIGHTS])

    def update_algorithm_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of Worker.algo, change its value to
        `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Algorithm attribute name
        """

        self.local_worker.update_algorithm_parameter(parameter_name, new_parameter_value)
        for e in self.remote_workers:
            e.update_algorithm_parameter.remote(parameter_name, new_parameter_value)

        self.algo.update_algorithm_parameter(parameter_name, new_parameter_value)
        for e in self.col_workers.remote_workers():
            e.update_algorithm_parameter.remote(parameter_name, new_parameter_value)

    def save_model(self, fname):
        """
        Save current version of actor as a torch loadable checkpoint.

        Parameters
        ----------
        fname : str
            Filename given to the checkpoint.

        Returns
        -------
        save_name : str
            Path to saved file.
        """

        torch.save(self.local_worker.actor.state_dict(), fname + ".tmp")
        os.replace(fname + '.tmp', fname)
        save_name = fname + ".{}".format(self.local_worker.actor_version)
        copy2(fname, save_name)
        return save_name

    def stop(self):
        """Stop collecting data."""

        if hasattr(self, "collector"):
            self.collector.stopped = True
        self.local_worker.stop()
        for e in self.remote_workers:
            e.stop.remote()
            e.terminate_worker.remote()


class CollectorThread(threading.Thread):
    """
    This class receives data samples from the data collection workers and
    queues them into the data input_queue.

    Parameters
    ----------
    index_worker : int
        Index assigned to this worker.
    input_queue : queue.Queue
        Queue to store the data dicts received from data collection workers.
    local_worker : Worker
        Local worker that acts as a parameter server.
    remote_workers : list of Workers
        Set of workers collecting and sending rollouts.
    col_fraction_workers : float
        Minimum fraction of samples required to stop if collection is
        synchronously coordinated and most workers have finished their
        collection task.
    col_communication : str
        Communication coordination pattern for data collection.
    col_execution : str
        Execution patterns for data collection.
    broadcast_interval : int
        After how many central updates, model weights should be broadcasted to
        remote collection workers.

    Attributes
    ----------
    stopped : bool
        Whether or not the thread in running.
    queue : queue.Queue
        Queue to store the data dicts received from data collection workers.
    index_worker : int
        Index assigned to this worker.
    local_worker : CWorker
        col_workers local worker.
    remote_workers : List of CWorker's
        col_workers remote data collection workers.
    num_workers : int
        Number of collection remote workers.
    broadcast_interval : int
        After how many collection step model weights should be broadcasted to
        remote collection workers.
    num_sent_since_broadcast : int
        Number of data dicts received since last model weights were broadcasted.
    """

    def __init__(self,
                 index_worker,
                 local_worker,
                 remote_workers,
                 col_fraction_workers=1.0,
                 col_communication=prl.SYNC,
                 col_execution=prl.CENTRAL,
                 broadcast_interval=1):

        self.index_worker = index_worker
        threading.Thread.__init__(self)

        self.stopped = False
        self.queue = queue.SimpleQueue()
        self.col_execution = col_execution
        self.col_communication = col_communication
        self.broadcast_interval = broadcast_interval
        self.fraction_workers = col_fraction_workers

        self.local_worker = local_worker
        self.remote_workers = remote_workers
        self.num_remote_workers = len(self.remote_workers)
        self.num_collection_workers = 1 if self.num_remote_workers == 0 else self.num_remote_workers

        # Counters
        self.num_sent_since_broadcast = 0

        if col_execution == prl.CENTRAL and col_communication == prl.SYNC:
            pass

        elif col_execution == prl.CENTRAL and col_communication == prl.ASYNC:
            if self.local_worker.envs_train: self.start()  # Start CollectorThread

        elif col_execution == prl.PARALLEL and col_communication == prl.SYNC:
            pass

        elif col_execution == prl.PARALLEL and col_communication == prl.ASYNC:
            self.pending_tasks = {}
            self.broadcast_new_weights()
            for w in self.remote_workers:
                    future = w.collect_data.remote()
                    self.pending_tasks[future] = w
            self.start()

    def run(self):
        while not self.stopped:
            # First, collect data
            self.step()
            # Then, update counter and broadcast weights to worker if necessary
            self.num_sent_since_broadcast += 1
            if self.should_broadcast():
                self.broadcast_new_weights()

    def step(self):
        """
        Collects data from remote workers and puts it in the GWorker queue.
        """

        if self.col_execution == prl.CENTRAL and self.col_communication == prl.SYNC:

            rollouts = self.local_worker.collect_data(listen_to=["sync"], data_to_cpu=False)
            rollouts = unpack(rollouts) if type(rollouts) == str else rollouts
            self.queue.put(rollouts)
            while self.queue.qsize() >= max_queue_size:
                time.sleep(0.5)

        elif self.col_execution == prl.CENTRAL and self.col_communication == prl.ASYNC:

            rollouts = self.local_worker.collect_data(data_to_cpu=False)
            rollouts = unpack(rollouts) if type(rollouts) == str else rollouts
            self.queue.put(rollouts)
            while self.queue.qsize() >= max_queue_size:
                time.sleep(0.5)

        elif self.col_execution == prl.PARALLEL and self.col_communication == prl.SYNC:

            # Start data collection in all workers
            worker_key = "worker_{}".format(self.index_worker)
            broadcast_message(worker_key, b"start-continue")

            pending_samples = [e.collect_data.remote(
                listen_to=["sync", worker_key]) for e in self.remote_workers]

            if self.fraction_workers < 1.0:
                # Keep checking how many workers have finished until
                # percent% are ready
                samples_ready = []
                while len(samples_ready) < (self.num_remote_workers * self.fraction_workers):
                    samples_ready, samples_not_ready = ray.wait(
                        pending_samples, num_returns=len(pending_samples), timeout=0.001)

                # Send stop message to the workers
                broadcast_message(worker_key, b"stop")

            # Compute model updates
            for r in pending_samples:
                rollouts = ray_get_and_free(r)
                rollouts = unpack(rollouts) if type(rollouts) == str else rollouts
                self.queue.put(rollouts)
                while self.queue.qsize() >= max_queue_size:
                    time.sleep(0.5)

        elif self.col_execution == prl.PARALLEL and self.col_communication == prl.ASYNC:

            # Wait for first worker to finish
            assert len(list(self.pending_tasks.keys())) == len(self.remote_workers)

            wait_results = ray.wait(list(self.pending_tasks.keys()))
            future = wait_results[0][0]
            w = self.pending_tasks.pop(future)

            # Retrieve rollouts and add them to queue
            rollouts = ray_get_and_free(future)
            rollouts = unpack(rollouts) if type(rollouts) == str else rollouts
            self.queue.put(rollouts)

            while self.queue.qsize() >= max_queue_size:
                time.sleep(0.5)

            # Schedule a new collection task
            future = w.collect_data.remote()
            self.pending_tasks[future] = w

    def should_broadcast(self):
        """Returns whether broadcast() should be called to update weights."""
        return self.num_sent_since_broadcast >= self.broadcast_interval

    def broadcast_new_weights(self):
        """Broadcast a new set of weights from the local worker."""
        if self.num_remote_workers > 0:
            latest_weights = ray.put({
                prl.VERSION: self.local_worker.actor_version,
                prl.WEIGHTS: self.local_worker.get_weights()})
            for e in self.remote_workers:
                e.set_weights.remote(latest_weights)
            self.num_sent_since_broadcast = 0





from pytorchrl.agent.algorithms.on_policy.ppo import PPO
from pytorchrl.agent.algorithms.on_policy.a2c import A2C
from pytorchrl.agent.algorithms.on_policy.rnd_ppo import RND_PPO

from pytorchrl.agent.algorithms.off_policy.ddpg import DDPG
from pytorchrl.agent.algorithms.off_policy.mpo import MPO
from pytorchrl.agent.algorithms.off_policy.ddqn import DDQN
from pytorchrl.agent.algorithms.off_policy.sac import SAC
from pytorchrl.agent.algorithms.off_policy.td3 import TD3

from pytorchrl.agent.algorithms.model_based.mpc_rs import MPC_RS
from pytorchrl.agent.algorithms.model_based.mpc_cem import MPC_CEM
from pytorchrl.agent.algorithms.model_based.mpc_pddm import MPC_PDDM



import os
import animalai
from animalai.envs.gym.environment import AnimalAIGym
from pytorchrl.envs.common import FrameStack, FrameSkip, DelayedReward
from pytorchrl.envs.animal_olympics.wrappers import RetroEnv, FilterActionEnv, LabAnimal, RewardShaping


def animal_train_env_factory(
        index_col_worker, index_grad_worker, index_env=0, frame_skip=0, frame_stack=1,
        arenas_dir=None, reduced_actions=True, reward_shape=True, exe_path=None,
        inference=False, reward_delay=1, id_offset=1):
    """
    Create train Animal Olympics Unity3D environment.

    Parameters
    ----------
     index_col_worker : int
        Index of the collection worker running this environment.
    index_grad_worker : int
        Index of the gradient worker running the collection worker running this environment.
    index_env : int
        Index of this environment withing the vector of environments.
    frame_skip : int
        Return only every `frame_skip`-th observation.
    frame_stack : int
        Observations composed of last `frame_stack` frames stacked.
    arenas_dir:
        path to dir containing arenas .yaml files.
    reduced_actions : bool
        Whether or not to use the action wrapper to reduce the number of available actions.
    reward_shape : bool
        Whether or not to use the reward shape wrapper.
    exe_path : str
        Path to obstacle environment executable.
    reward_delay : int
        Only return accumulated reward every `reward_delay` steps to simulate sparse reward environment.
    inference : bool
        Whether or not to render the environment frames in real time.
    id_offset : int
        offset added to worker_id to avoid collisions with other runs in the same machine.

    Returns
    -------
    env : gym.Env
        Train environment.
    """
    if 'DISPLAY' not in os.environ.keys():
        os.environ['DISPLAY'] = ':0'

    if exe_path:
        exe = exe_path
    else:
        exe = os.path.join(os.path.dirname(animalai.__file__), '../../examples/env/AnimalAI')

    id = id_offset + index_grad_worker * 10000 + 1000 * index_col_worker + index_env * 100

    env = AnimalAIGym(
        exe,
        worker_id=id,
        seed=id,
        n_arenas=1,
        arenas_configurations=None,
        inference=inference)

    env = RetroEnv(env)

    if reduced_actions:
        env = FilterActionEnv(env)

    env = LabAnimal(env, arenas_dir)

    if reward_shape:
        env = RewardShaping(env)

    if frame_skip > 0:
        env = FrameSkip(env, skip=frame_skip)

    if frame_stack > 1:
        env = FrameStack(env, k=frame_stack)

    if reward_delay > 1:
        env = DelayedReward(env, delay=reward_delay)

    return env


def animal_test_env_factory(
        index_col_worker, index_grad_worker, index_env=0, frame_skip=0, frame_stack=1,
        arenas_dir=None, reduced_actions=True, reward_shape=False, exe_path=None,
        inference=False, reward_delay=1, id_offset=1):
    """
    Create train Animal Olympics Unity3D environment.

    Parameters
    ----------
     index_col_worker : int
        Index of the collection worker running this environment.
    index_grad_worker : int
        Index of the gradient worker running the collection worker running this environment.
    index_env : int
        Index of this environment withing the vector of environments.
    frame_skip : int
        Return only every `frame_skip`-th observation.
    frame_stack : int
        Observations composed of last `frame_stack` frames stacked.
    arenas_dir:
        path to dir containing arenas .yaml files.
    reduced_actions : bool
        Whether or not to use the action wrapper to reduce the number of available actions.
    reward_shape : bool
        Whether or not to use the reward shape wrapper.
    exe_path : str
        Path to obstacle environment executable.
    reward_shape : bool
        Whether or not to render the environment in real time.
    reward_delay : int
        Only return accumulated reward every `reward_delay` steps to simulate sparse reward environment.
    inference : bool
        Whether or not to render the environment frames in real time.
    id_offset : int
        offset added to worker_id to avoid collisions with other runs in the same machine.

    Returns
    -------
    env : gym.Env
        Train environment.
    """
    if 'DISPLAY' not in os.environ.keys():
        os.environ['DISPLAY'] = ':0'

    if exe_path:
        exe = exe_path
    else:
        exe = os.path.join(os.path.dirname(animalai.__file__), '../../examples/env/AnimalAI')

    id = id_offset + index_grad_worker * 10000 + 1000 * index_col_worker + index_env * 100
    env = AnimalAIGym(
        exe,
        worker_id=id,
        seed=id,
        n_arenas=1,
        arenas_configurations=None,
        inference=inference)

    env = RetroEnv(env)

    if reduced_actions:
        env = FilterActionEnv(env)

    env = LabAnimal(env, arenas_dir)

    if reward_shape:
        env = RewardShaping(env)

    if frame_skip > 0:
        env = FrameSkip(env, skip=frame_skip)

    if frame_stack > 1:
        env = FrameStack(env, k=frame_stack)

    if reward_delay > 1:
        env = DelayedReward(env, delay=reward_delay)

    return env



import torch
import torch.nn as nn
from abc import ABC, abstractmethod
from pytorchrl.agent.actors.utils import partially_load_checkpoint


class Actor(nn.Module, ABC):
    """
    Base class for all Actors.

    Parameters
    ----------
    device: torch.device
        CPU or specific GPU where class computations will take place.
    input_space : gym.Space
        Environment observation space.
    action_space : gym.Space
        Environment action space.
    checkpoint : str
        Path to a previously trained Actor checkpoint to be loaded.
    """
    def __init__(self,
                 device,
                 input_space,
                 action_space,
                 checkpoint=None,
                 *args):

        super(Actor, self).__init__()
        self.device = device
        self.checkpoint = checkpoint
        self.input_space = input_space
        self.action_space = action_space

    @classmethod
    @abstractmethod
    def create_factory(
            cls,
            device,
            input_space,
            action_space,
            checkpoint=None,
            *args):
        """
        Returns a function that creates actor critic instances.

        Parameters
        ----------
        device: torch.device
            CPU or specific GPU where class computations will take place.
        input_space : gym.Space
            Environment observation space.
        action_space : gym.Space
            Environment action space.
        checkpoint : str
            Path to a previously trained Actor checkpoint to be loaded.

        Returns
        -------
        create_actor_instance : func
            creates a new Actor class instance.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def is_recurrent(self, *args):
        """Returns True if the actor network are recurrent."""
        raise NotImplementedError

    @property
    @abstractmethod
    def recurrent_hidden_state_size(self):
        """Size of policy recurrent hidden state"""
        raise NotImplementedError

    @abstractmethod
    def actor_initial_states(self, obs, *args):
        """
        Returns all policy inputs to predict the environment initial action.

        Parameters
        ----------
        obs : torch.tensor
            Initial environment observation.

        Returns
        -------
        obs : torch.tensor
            Initial environment observation.
        rhs : torch.tensor
            Initial recurrent hidden state.
        done : torch.tensor
            Initial done tensor, indicating the environment is not done.
        """
        raise NotImplementedError

    def try_load_from_checkpoint(self):
        """Load weights from previously saved checkpoint."""
        try:
            if isinstance(self.checkpoint, str):
                print("Loading all model weight from {}".format(self.checkpoint))
                self.load_state_dict(torch.load(self.checkpoint, map_location=self.device))
            elif isinstance(self.checkpoint, dict):
                for submodule, checkpoint in self.checkpoint.items():
                    print("Loading {} model weight from {}".format(submodule, self.checkpoint[submodule]))
                    partially_load_checkpoint(self, submodule, checkpoint, map_location=self.device)
            else:
                print("Training model from scratch")
        except Exception:
            print("Error when trying to load checkpoint!")



import gym
import random
import numpy as np
import rdkit as Chem
from gym import spaces
from collections import defaultdict, deque
from reinvent_chemistry import Conversions
from reinvent_chemistry.library_design import BondMaker, AttachmentPoints
from reinvent_chemistry.library_design.reaction_filters.reaction_filter import ReactionFilter, \
    ReactionFilterConfiguration
from pytorchrl.envs.generative_chemistry.diversity_filter.no_filter_with_penalty import NoFilterWithPenalty


class GenChemEnv(gym.Env):
    """Custom Environment for Generative Chemistry RL."""

    metadata = {'render.modes': ['human']}

    def __init__(self, scoring_function, vocabulary, scaffolds, randomize_scaffolds=False, max_length=200,
                 reactions=[]):
        super(GenChemEnv, self).__init__()

        self.num_episodes = 0
        self.reactions = reactions
        self.scaffolds = scaffolds
        self.vocabulary = vocabulary
        self.max_length = max_length
        self.current_episode_length = 0
        self.scoring_function = scoring_function
        self.randomize_scaffolds = randomize_scaffolds
        self.running_mean_valid_smiles = deque(maxlen=1)

        self._bond_maker = BondMaker()
        self._conversion = Conversions()
        self._attachment_points = AttachmentPoints()

        # Check maximum possible scaffold length
        self.max_scaffold_length = max([len(self.select_scaffold()) for _ in range(1000)])

        # Define action and observation space
        self.action_space = gym.spaces.Discrete(len(self.vocabulary.decoration_vocabulary))
        scaffold_space = gym.spaces.Discrete(len(self.vocabulary.scaffold_vocabulary))
        scaffold_length = gym.spaces.Discrete(self.max_scaffold_length)
        decoration_space = gym.spaces.Discrete(len(self.vocabulary.decoration_vocabulary))
        decoration_length = gym.spaces.Discrete(self.max_length)
        full_decoration_space = gym.spaces.Discrete(len(self.vocabulary.decoration_vocabulary))
        full_decoration_length = gym.spaces.Discrete(self.max_length)

        # Ugly hack
        scaffold_space._shape = (self.max_scaffold_length,)
        # full_decoration_space._shape = (self.max_length,)
        decoration_space._shape = (1,)

        self.observation_space = gym.spaces.Dict({
            "context": scaffold_space,
            "context_length": scaffold_length,
            "obs": decoration_space,
            "obs_length": decoration_length,
            # "full_obs": full_decoration_space,
            # "full_obs_length": full_decoration_length,
        })

        # Reaction Filters
        reaction_filter_conf = {"type": "selective", "reactions": reactions}
        reaction_filter_conf = ReactionFilterConfiguration(
            type=reaction_filter_conf["type"],
            reactions=reaction_filter_conf["reactions"],
            reaction_definition_file=None)
        self.reaction_filter = ReactionFilter(reaction_filter_conf)

        # Diversity Filter: penalizes the score by 0.5 if a previously seen compound is proposed.
        self.diversity_filter = NoFilterWithPenalty()

    def step(self, action):
        """Execute one time step within the environment"""

        info = {}
        self.current_decoration_length += 1
        action = "$" if self.current_decoration_length == self.max_length - 1 else \
            self.vocabulary.decode_decoration_token(action)
        self.current_decoration += action

        if action != "$":  # If character is not $, return 0.0 reward
            reward = 0.0
            done = False

        else:  # if action is $, evaluate molecule

            decorated_smile, molecule = self.join_scaffold_and_decorations(
                self.vocabulary.decode_scaffold(self.padded_scaffold),
                self.vocabulary.remove_start_and_end_tokens(self.current_decoration)
            )

            # Compute score
            score = self.scoring_function(decorated_smile)

            # Sanity check
            assert isinstance(score, dict), "scoring_function has to return a dict"
            assert "score" in score.keys() or "reward" in score.keys(), \
                "scoring_function outputs requires at lest the keyword ´score´ or ´reward´"

            # Apply reaction filters
            score.update({"reaction_scores": 0.0})
            if molecule:
                self.apply_reaction_filters(molecule, score)

            # Get reward
            reward = score["reward"] if "reward" in score.keys() else score["score"]

            # Adjust reward with diversity filter
            reward = self.diversity_filter.update_score(reward, decorated_smile)

            # If score contain field "Valid", update counter
            if "valid_smile" in score.keys():
                valid = score["valid_smile"]
                self.running_mean_valid_smiles.append(1.0) if valid else \
                    self.running_mean_valid_smiles.append(0.0)

            # Update molecule
            info.update({"molecule": decorated_smile or "invalid_smile"})

            # Update info with remaining values
            info.update(score)

            # Update valid smiles tracker
            info.update({"valid_smile": float((sum(self.running_mean_valid_smiles) / len(
                self.running_mean_valid_smiles)) * 100 if len(self.running_mean_valid_smiles) != 0.0 else 0.0)})

            done = True

        self.padded_current_decoration[self.current_decoration_length - 1] = \
            self.vocabulary.encode_decoration_token(action)

        next_obs = {
            "context": self.padded_scaffold,
            "context_length": np.array(self.scaffold_length),
            "obs": np.array(self.vocabulary.encode_decoration_token(action)).reshape(1),
            "obs_length": np.array(1),
            # "full_obs":  self.padded_current_decoration,
            # "full_obs_length": np.array(self.current_decoration_length),
        }

        return next_obs, reward, done, info

    def reset(self):
        """
        Reset the state of the environment to an initial state.
        Return padded base molecule to match length `obs_length`.
        """
        self.num_episodes += 1

        self.scaffold_length = np.Inf
        while self.scaffold_length > self.max_scaffold_length:
            self.scaffold = self.select_scaffold()
            self.scaffold_length = len(self.scaffold)

        self.padded_scaffold = self.vocabulary.encode_scaffold_token('<pad>') * np.ones(self.max_scaffold_length)
        self.padded_scaffold[0:self.scaffold_length] = self.scaffold

        self.current_decoration = "^"
        self.padded_current_decoration = self.vocabulary.encode_decoration_token('<pad>') * np.ones(self.max_length)
        self.padded_current_decoration[0] = self.vocabulary.encode_decoration_token(self.current_decoration)
        self.current_decoration_length = 1

        # "decoration":  self.padded_current_decoration,
        # "decoration_length": np.array(self.current_decoration_length),

        obs = {
            "context": self.padded_scaffold,
            "context_length": np.array(self.scaffold_length),
            "obs": np.array(self.vocabulary.encode_decoration_token(self.current_decoration)).reshape(1),
            "obs_length": np.array(1),
            # "full_obs":  self.padded_current_decoration,
            # "full_obs_length": np.array(self.current_decoration_length),
        }

        return obs

    def render(self, mode='human'):
        """Render the environment to the screen"""

        print(f'Current Molecule: {self.current_molecule}')
        print(f'Vocabulary: {self.vocabulary._tokens}')

    def select_scaffold(self):
        scaffold = random.choice(self.scaffolds)
        if self.randomize_scaffolds and len(self.reactions) == 0:
            mol = self._conversion.smile_to_mol(scaffold)
            scaffold = self._bond_maker.randomize_scaffold(mol)  # randomize
        scaffold = self._attachment_points.remove_attachment_point_numbers(scaffold)
        scaffold = self.vocabulary.encode_scaffold(scaffold)
        return scaffold

    def join_scaffold_and_decorations(self, scaffold, decorations):
        scaffold = self._attachment_points.add_attachment_point_numbers(scaffold, canonicalize=False)
        molecule = self._bond_maker.join_scaffolds_and_decorations(scaffold, decorations)
        smile = self._conversion.mol_to_smiles(molecule) if molecule else None
        return smile, molecule

    def apply_reaction_filters(self, mol, final_score):
        reaction_scores = [self.reaction_filter.evaluate(mol) if mol else 0.0]
        reward = final_score["reward"] if "reward" in final_score.keys() else final_score["score"]
        final_score["reward"] = float(reward * np.array(reaction_scores))
        final_score["reaction_scores"] = float(np.array(reaction_scores))
        return final_score



import os
import yaml
import torch
import shutil
import argparse
from torch._six import inf


def cleanup_log_dir(log_dir):
    """
    Create log directory and remove old files.

    Parameters
    ----------
    log_dir : str
        Path to log directory.
    """
    try:
        shutil.rmtree(os.path.join(log_dir))
    except Exception:
        print("Unable to cleanup log_dir...")
    os.makedirs(log_dir, exist_ok=True)


class LoadFromFile(argparse.Action):
    # parser.add_argument('--file', type=open, action=LoadFromFile)
    def __call__(self, parser, namespace, values, option_string=None):
        if values.name.endswith('yaml') or values.name.endswith('yml'):
            with values as f:
                config = yaml.load(f, Loader=yaml.FullLoader)
            for key in config.keys():
                if key not in namespace:
                    raise ValueError(f'Unknown argument in config file: {key}')
            namespace.__dict__.update(config)
        else:
            raise ValueError('Configuration file must end with yaml or yml')


def save_argparse(args, filename, exclude=None):
    if filename.endswith('yaml') or filename.endswith('yml'):
        if isinstance(exclude, str):
            exclude = [exclude, ]
        args = args.__dict__.copy()
        for exl in exclude:
            del args[exl]
        yaml.dump(args, open(filename, 'w'))
    else:
        raise ValueError("Configuration file should end with yaml or yml")


class RunningMeanStd:
    """Class to keep track on the running mean and variance of tensors batches."""

    def __init__(self, epsilon=1e-4, shape=(), device=torch.device("cpu")):
        self.mean = torch.zeros(shape, dtype=torch.float64).to(device)
        self.var = torch.ones(shape, dtype=torch.float64).to(device)
        self.count = epsilon

    def update(self, x):
        batch_mean = torch.mean(x, dim=0)
        batch_var = torch.var(x, dim=0)
        batch_count = x.shape[0]
        self.update_from_moments(batch_mean, batch_var, batch_count)

    def update_from_moments(self, batch_mean, batch_var, batch_count):
        delta = batch_mean - self.mean
        tot_count = self.count + batch_count
        new_mean = self.mean + delta * batch_count / tot_count
        m_a = self.var * self.count
        m_b = batch_var * batch_count
        M2 = m_a + m_b + torch.square(delta) * self.count * batch_count / tot_count
        new_var = M2 / tot_count
        new_count = tot_count
        self.mean, self.var, self.count = new_mean, new_var, new_count


def clip_grad_norm_(parameters, norm_type: float = 2.0):
    """
    This is the official clip_grad_norm implemented in pytorch but the max_norm part has been removed.
    https://github.com/pytorch/pytorch/blob/52f2db752d2b29267da356a06ca91e10cd732dbc/torch/nn/utils/clip_grad.py#L9
    """
    if isinstance(parameters, torch.Tensor):
        parameters = [parameters]
    parameters = [p for p in parameters if p.grad is not None]
    norm_type = float(norm_type)
    if len(parameters) == 0:
        return torch.tensor(0.)
    device = parameters[0].grad.device
    if norm_type == inf:
        total_norm = max(p.grad.detach().abs().max().to(device) for p in parameters)
    else:
        total_norm = torch.norm(
            torch.stack([torch.norm(p.grad.detach(), norm_type).to(device) for p in parameters]), norm_type)
    return total_norm



import os
import glob
import copy
import uuid
import torch
import numpy as np
from copy import deepcopy
from collections import defaultdict

import pytorchrl as prl
from pytorchrl.utils import RunningMeanStd
from pytorchrl.agent.actors.utils import partially_load_checkpoint
from pytorchrl.agent.storages.on_policy.ppod2_buffer import PPOD2Buffer as B

MIN_BUFFER_SIZE = 10


class PPOD2RebelBuffer(B):
    """
    Storage class for a modified version of the PPO+D algorithm.

    This version of the PPO+D buffer uses intrinsic rewards (if available) instead of value predictions to rank
    potentially interesting demos. Additionally, this version does not differentiate between human demos and agent
    demos, only between reward demos (independently of their origin) and intrinsic demos. If no intrinsic rewards
    are available, only reward demos are replayed.

    Parameters
    ----------
    size : int
        Storage capacity along time axis.
    device: torch.device
        CPU or specific GPU where data tensors will be placed and class
        computations will take place. Should be the same device where the
        actor model is located.
    actor : Actor
        Actor class instance.
    algorithm : Algorithm
        Algorithm class instance.
    envs : VecEnv
        Vector of experiments instance.
    reward_predictor_factory : func
        Method to create the reference value function.
    rho : float
        PPO+D rho parameter.
    phi : float
        PPO+D phi parameter.
    gae_lambda : float
        GAE lambda parameter.
    error_threshold : float
        Minimum value prediction error. Below error_threshold, reward sign is flipped.
    alpha : float
        PPO+D alpha parameter
    initial_reward_threshold : float
        initial intrinsic to use as reward threshold for new reward_demos.
    initial_reward_demos_dir : str
        Path to directory containing agent initial demonstrations.
    supplementary_demos_dir : str
        Path to a directory where additional demos can be added after training has started.
        these demos will be incorporated into the buffer as bonus agent demos.
    target_reward_demos_dir : str
        Path to directory where best reward demonstrations should be saved.
    num_reward_demos_to_save : int
        Number of top reward reward_demos to save.
    initial_int_demos_dir : str
        Path to directory containing intrinsic initial demonstrations.
    target_int_demos_dir : str
        Path to directory where best intrinsic demonstrations should be saved.
    num_int_demos_to_save : int
        Number of top intrinsic reward_demos to save.
    total_buffer_demo_capacity : int
        Maximum number of reward_demos to keep between reward and intrinsic reward_demos.
    save_demos_every : int
        Save top reward_demos every  `save_demo_frequency`th data collection.
    demo_dtypes : dict
        data types to use for the reward_demos.
    """

    # Accepted data fields. Inserting other fields will raise AssertionError
    on_policy_data_fields = prl.OnPolicyDataKeys

    # Data tensors to collect for each reward_demos
    demos_data_fields = prl.DemosDataKeys

    def __init__(self, size, device, actor, algorithm, envs, reward_predictor_factory=None,
                 reward_predictor_net_kwargs={}, restart_reward_predictor_net=None, rho=0.05, phi=0.05, gae_lambda=0.95,
                 alpha=10, total_buffer_demo_capacity=50, initial_reward_threshold=None, initial_reward_demos_dir=None,
                 supplementary_demos_dir=None, target_reward_demos_dir=None, num_reward_demos_to_save=None,
                 initial_int_demos_dir=None, target_int_demos_dir=None, num_int_demos_to_save=None,
                 save_demos_every=10, demo_dtypes={prl.OBS: np.float32, prl.ACT: np.float32,  prl.REW: np.float32}):

        super(PPOD2RebelBuffer, self).__init__(
            size, device, actor, algorithm, envs, rho, phi, gae_lambda, alpha, total_buffer_demo_capacity,
            initial_reward_threshold, initial_reward_demos_dir, supplementary_demos_dir, target_reward_demos_dir,
            num_reward_demos_to_save, initial_int_demos_dir, target_int_demos_dir, num_int_demos_to_save,
            save_demos_every, demo_dtypes)

        # Create target demo dirs
        os.makedirs(target_reward_demos_dir, exist_ok=True)

        # Create general value model and move it to device
        if reward_predictor_factory:
            model = torch.nn.Module()
            model.reward_predictor = reward_predictor_factory(**reward_predictor_net_kwargs).to(self.device)
            model.error_threshold = torch.nn.parameter.Parameter(
                data=torch.tensor(-1000000, dtype=torch.float32), requires_grad=False)
            if restart_reward_predictor_net:
                partially_load_checkpoint(
                    model, "reward_predictor", restart_reward_predictor_net, map_location=self.device)
                partially_load_checkpoint(
                    model, "error_threshold", restart_reward_predictor_net, map_location=self.device)

            # Freeze general value model with respect to optimizers
            for p in model.reward_predictor.parameters():
                p.requires_grad = False

            self.actor.predictor = model

        else:
            self.actor.predictor = None

        # Define reward and error threshold
        self.reward_threshold = self.initial_reward_threshold = initial_reward_threshold or 0.0

    @classmethod
    def create_factory(cls, size, reward_predictor_factory=None, reward_predictor_net_kwargs={},
                       restart_reward_predictor_net=None, rho=0.05, phi=0.05, gae_lambda=0.95, alpha=10,
                       total_buffer_demo_capacity=50, initial_reward_threshold=None, initial_reward_demos_dir=None,
                       supplementary_demos_dir=None, target_reward_demos_dir=None, num_reward_demos_to_save=None,
                       initial_int_demos_dir=None, target_int_demos_dir=None, num_int_demos_to_save=None,
                       save_demos_every=10, demo_dtypes={prl.OBS: np.float32, prl.ACT: np.float32, prl.REW: np.float32}):
        """
        Returns a function that creates PPOD2RebelBuffer instances.

        Parameters
        ----------
        size : int
            Storage capacity along time axis.
        rho : float
            PPO+D rho parameter.
        phi : float
            PPO+D phi parameter.
        gae_lambda : float
            GAE lambda parameter.
        alpha : float
            PPO+D alpha parameter
        total_buffer_demo_capacity : int
            Maximum number of reward_demos to keep between reward and intrinsic reward_demos.
            initial_reward_threshold : float
        initial_reward_threshold : float
            initial intrinsic to use as reward threshold for new reward_demos.
        initial_reward_demos_dir : str
            Path to directory containing agent initial demonstrations.
        supplementary_demos_dir : str
            Path to a directory where additional demos can be added after training has started.
            these demos will be incorporated into the buffer as bonus agent demos.
        target_reward_demos_dir : str
            Path to directory where best reward demonstrations should be saved.
        num_reward_demos_to_save : int
            Number of top reward reward_demos to save.
        initial_int_demos_dir : str
            Path to directory containing intrinsic initial demonstrations.
        target_int_demos_dir : str
            Path to directory where best intrinsic demonstrations should be saved.
        num_int_demos_to_save : int
            Number of top intrinsic reward_demos to save.
        total_buffer_demo_capacity : int
            Maximum number of reward_demos to keep between reward and intrinsic reward_demos.
        save_demos_every : int
            Save top reward_demos every  `save_demo_frequency`th data collection.
        demo_dtypes : dict
            data types to use for the reward_demos.

        Returns
        -------
        create_buffer_instance : func
            creates a new PPOD2RebelBuffer class instance.
        """
        def create_buffer_instance(device, actor, algorithm, envs):
            """Create and return a PPOD2RebelBuffer instance."""
            return cls(size, device, actor, algorithm, envs, reward_predictor_factory, reward_predictor_net_kwargs,
                       restart_reward_predictor_net, rho, phi, gae_lambda, alpha, total_buffer_demo_capacity,
                       initial_reward_threshold, initial_reward_demos_dir, supplementary_demos_dir,
                       target_reward_demos_dir, num_reward_demos_to_save, initial_int_demos_dir, target_int_demos_dir,
                       num_int_demos_to_save, save_demos_every, demo_dtypes)
        return create_buffer_instance

    def init_tensors(self, sample):
        """
        Lazy initialization of data tensors from a sample.

        Parameters
        ----------
        sample : dict
            Data sample (containing all tensors of an environment transition)"""

        super(PPOD2RebelBuffer, self).init_tensors(sample)

        # Add cumulative rewards tensor to self.data
        self.data["CumRew"] = torch.zeros_like(self.data[prl.REW])

    def before_gradients(self):
        """Before updating actor policy model, compute returns and advantages."""
        if self.actor.predictor:
            self.compute_cumulative_rewards()
            self.apply_rebel_logic()
        super(PPOD2RebelBuffer, self).before_gradients()

    def after_gradients(self, batch, info):
        """
        After updating actor policy model, make sure self.step is at 0.

        Parameters
        ----------
        batch : dict
            Data batch used to compute the gradients.
        info : dict
            Additional relevant info from gradient computation.

        Returns
        -------
        info : dict
            info dict updated with relevant info from Storage.
        """

        info = super(PPOD2RebelBuffer, self).after_gradients(batch, info)

        info["Algorithm"].update({
            "max_modified_reward": self.data[prl.REW].max().item(),
            "min_modified_reward": self.data[prl.REW].min().item(),
            "avg_modified_reward": self.data[prl.REW].mean().item(),
        })

        return info

    def apply_rebel_logic(self):
        ref_value = self.predict_reference_value()
        self.modify_rewards(ref_value)

    def compute_cumulative_rewards(self):
        """Compute cumulative episode rewards and also returns."""
        length = self.step - 1 if self.step != 0 else self.max_size
        self.data[prl.RET][length].copy_(self.data[prl.VAL][length])
        self.data["CumRew"].copy_(self.data[prl.REW])
        for step in reversed(range(length)):
            self.data[prl.RET][step] = (self.data[prl.RET][step + 1] * self.algo.gamma * (
                    1.0 - self.data[prl.DONE][step + 1]) + self.data[prl.REW][step])
        for step in range(length + 1):
            self.data["CumRew"][step] += self.data["CumRew"][step - 1] * (1.0 - self.data[prl.DONE][step])

    def predict_reference_value(self):
        """Generate value predictions with the reference value network."""
        return self.actor.predictor.reward_predictor(
            self.data[prl.OBS].view(-1, *self.data[prl.OBS].shape[2:])
        ).view(self.data[prl.DONE].shape)

    def modify_rewards(self, ref_value):
        """Flip the sign of the rewards with lower reference value prediction than self.actor.error_threshold."""
        errors = torch.abs(ref_value - self.data[prl.RET])
        mask = (errors < float(self.actor.predictor.error_threshold)) * (self.data["CumRew"] >= float(
            self.reward_threshold)) * (self.data[prl.REW] > 0.0)
        self.data[prl.REW][mask] *= -1

    def validate_demo(self, demo):
        """
        Verify demo has high value prediction error for rewarded states with
        cumulative reward higher than self.reward_threshold - 1.
        """

        if not self.actor.predictor.reward_predictor:
            return True

        # Compute demo cumulative rewards
        cumulative_rewards = np.copy(demo[prl.REW])
        for step in range(1, len(cumulative_rewards)):
            cumulative_rewards[step] += cumulative_rewards[step - 1]
        cumulative_rewards = cumulative_rewards[self.frame_stack - 1:]

        # Get demo rewards
        rewards = np.copy(demo[prl.REW])[self.frame_stack - 1:]

        # Define mask to get only final states
        mask1 = np.logical_and(cumulative_rewards > self.initial_reward_threshold, rewards != 0)  # after thresh
        mask2 = np.logical_and(cumulative_rewards <= self.initial_reward_threshold, rewards != 0)  # from 0 to thresh

        # Create stacked observations
        stacked_obs = []
        for start in range(self.frame_stack):
            end = - (self.frame_stack - 1 - start)
            if end == 0:
                end = None
            stacked_obs.append(demo[prl.OBS][start:end])
        stacked_obs = np.concatenate(stacked_obs, axis=1)

        # Compute reward prediction for rewarded states with cumulative rewards < self.reward_threshold
        reward_preds = self.actor.predictor.reward_predictor(torch.tensor(stacked_obs).to(self.device)).cpu().numpy()

        # Verify all predicted errors are higher than self.error_threshold at the end and lower at the beginning
        validation1 = (np.abs(reward_preds[mask1] - rewards[mask1]) > float(self.actor.predictor.error_threshold)).all()
        validation2 = (np.abs(reward_preds[mask2] - rewards[mask2]) < float(self.actor.predictor.error_threshold)).all()
        validation = validation1 and validation2

        return validation

    def track_potential_demos(self, sample):
        """ Tracks current episodes looking for potential agent_demos """

        for i in range(self.num_envs):

            # Copy transition
            for tensor in self.demos_data_fields:
                if tensor in (prl.OBS, ):
                    self.potential_demos["env{}".format(i + 1)][tensor].append(copy.deepcopy(
                        sample[tensor][i, -self.num_channels_obs:]).cpu().numpy().astype(self.demo_dtypes[tensor]))
                else:
                    self.potential_demos["env{}".format(i + 1)][tensor].append(
                        copy.deepcopy(sample[tensor][i]).cpu().numpy().astype(self.demo_dtypes[tensor]))

            # Update cumulative intrinsic reward
            if prl.IREW in sample.keys():

                # Track the cumulative sum of intrinsic rewards of the demo
                self.potential_demos_cumsum_int["env{}".format(i + 1)] += sample[prl.IREW][i].cpu().item()

                # Track the max intrinsic reward of the demo
                self.potential_demos_max_int["env{}".format(i + 1)] = max(
                    self.potential_demos_max_int["env{}".format(i + 1)], sample[prl.IREW][i].cpu().item())

            # Handle end of episode
            if sample[prl.DONE2][i] == 1.0:

                # Get candidate agent_demos
                potential_demo = {}
                for tensor in self.demos_data_fields:
                    potential_demo[tensor] = np.stack(self.potential_demos["env{}".format(i + 1)][tensor])
                    if tensor == prl.REW:
                        nonzero = np.flatnonzero(potential_demo[tensor] > 0.0)
                        last_reward = len(potential_demo[tensor]) if len(nonzero) == 0 else np.max(nonzero)

                # Compute accumulated reward
                episode_reward = potential_demo[prl.REW].sum()
                potential_demo["ID"] = str(uuid.uuid4())
                potential_demo["TotalReward"] = episode_reward
                potential_demo["DemoLength"] = potential_demo[prl.ACT].shape[0]

                # Consider candidate agent_demos for agent_demos reward
                if self.max_reward_demos > 0:

                    if episode_reward >= self.reward_threshold:

                        # Cut off data after last reward
                        for tensor in self.demos_data_fields:
                            potential_demo[tensor] = potential_demo[tensor][0:last_reward + 1]
                        potential_demo["DemoLength"] = potential_demo[prl.ACT].shape[0]

                        # Make sure new reward was previously unknown
                        valid = self.validate_demo(potential_demo)

                        # Add agent_demos to reward buffer
                        if valid:
                            self.reward_demos.append(potential_demo)

                        # Check if buffers are full
                        self.check_demo_buffer_capacity()

                        # Update reward_threshold.
                        self.reward_threshold = min([d["TotalReward"] for d in self.reward_demos])

                        # Update max demo reward
                        self.max_demo_reward = max([d["TotalReward"] for d in self.reward_demos])

                if self.max_intrinsic_demos > 0:

                    # Set potential demo cumulative intrinsic reward
                    episode_ireward = self.eta * self.potential_demos_max_int["env{}".format(i + 1)] + (1 - self.eta) * (
                            self.potential_demos_cumsum_int["env{}".format(i + 1)] / potential_demo["DemoLength"])
                    potential_demo["IntrinsicReward"] = episode_ireward

                    if episode_ireward >= self.intrinsic_threshold or \
                            len(self.intrinsic_demos) < self.max_intrinsic_demos:

                        # Add agent_demos to intrinsic buffer
                        self.intrinsic_demos.append(potential_demo)

                        # Check if buffer is full
                        self.check_demo_buffer_capacity()

                        # Update intrinsic_threshold
                        self.intrinsic_threshold = min([p["IntrinsicReward"] for p in self.intrinsic_demos])

                # Reset potential agent_demos dict
                for tensor in self.demos_data_fields:
                    self.potential_demos["env{}".format(i + 1)][tensor] = []
                    self.potential_demos_max_int["env{}".format(i + 1)] = 0.0
                    self.potential_demos_cumsum_int["env{}".format(i + 1)] = 0.0



from pytorchrl.agent.actors.reward_functions.gym_reward_functions import cartpole, pendulum, halfcheetah_mujoco, inverted_pendulum_mujoco
from pytorchrl.agent.actors.reward_functions.pybullet_reward_functions import halfcheetah_bullet
"""
Good Resource of reward functions: https://arxiv.org/pdf/1907.02057.pdf 
including gym envrionemnts as well as pybullet environments
"""


def get_reward_function(env_id):
    # gym envs
    if env_id == "CartPole-v0" or env_id == "CartPole-v1":
        return cartpole
    elif env_id == "Pendulum-v0" or env_id == "Pendulum-v1":
        return pendulum
    elif env_id == "InvertedPendulum-v2":
        return inverted_pendulum_mujoco
    elif env_id == "HalfCheetah-v2" or env_id == "HalfCheetah-v3":
        return halfcheetah_mujoco
    # pybullet envs
    elif env_id == "HalfCheetahBulletEnv-v0":
        return halfcheetah_bullet



import torch
import pytorchrl as prl
from pytorchrl.agent.storages.on_policy.vanilla_on_policy_buffer import VanillaOnPolicyBuffer as B


class VTraceBuffer(B):
    """
    Storage class for On-Policy algorithms with off-policy correction method
    V-trace (https://arxiv.org/abs/1506.02438).

    Parameters
    ----------
    size : int
        Storage capacity along time axis.
    device: torch.device
        CPU or specific GPU where data tensors will be placed and class
        computations will take place. Should be the same device where the
        actor model is located.
    envs : VecEnv
        Vector of environments instance.
    actor : Actor
        Actor class instance.
    algorithm : Algorithm
        Algorithm class instance.
    """

    # Data fields to store in buffer and contained in generated batches
    storage_tensors = prl.OnPolicyDataKeys

    def __init__(self, size, device, actor, algorithm, envs):

        super(VTraceBuffer, self).__init__(
            size=size,
            envs=envs,
            device=device,
            actor=actor,
            algorithm=algorithm)

    def before_gradients(self):
        """
        Before updating actor policy model, compute returns and advantages.
        """

        last_tensors = {}
        step = self.step if self.step != 0 else -1
        for k in (prl.OBS, prl.RHS, prl.DONE):
            if isinstance(self.data[k], dict):
                last_tensors[k] = {x: self.data[k][x][step] for x in self.data[k]}
            else:
                last_tensors[k] = self.data[k][step - 1]

        with torch.no_grad():
            _ = self.actor.get_action(last_tensors[prl.OBS], last_tensors[prl.RHS], last_tensors[prl.DONE])
            value_dict = self.actor.get_value(last_tensors[prl.OBS], last_tensors[prl.RHS], last_tensors[prl.DONE])
            next_value = value_dict.get("value_net1")
            next_rhs = value_dict.get("rhs")

        self.data[prl.RET][step].copy_(next_value)
        self.data[prl.VAL][step].copy_(next_value)

        if isinstance(next_rhs, dict):
            for x in self.data[prl.RHS]:
                self.data[prl.RHS][x][step].copy_(next_rhs[x])
        else:
            self.data[prl.RHS][step] = next_rhs
        self.compute_returns()

        self.compute_vtrace()

    @torch.no_grad()
    def get_updated_action_log_probs(self):
        """
        Computes new log probabilities of actions stored in `storage`
        according to current `actor` version. It also uses the current
        `actor` version to update the value predictions.
        """

        l, num_proc = self.data[prl.DONE].shape[0:2]
        l = self.step if self.step != 0 else self.max_size

        # Create batches without shuffling data
        batches = self.generate_batches(
            self.algo.num_mini_batch, self.algo.mini_batch_size,
            num_epochs=1, shuffle=False)

        # Obtain new value and log probability predictions
        new_val = []
        new_logp = []
        for batch in batches:
            obs, rhs, act, done = batch[prl.OBS], batch[prl.RHS], batch[prl.ACT], batch[prl.DONE]
            (logp, _, _) = self.actor.evaluate_actions(obs, rhs, done, act)
            val = self.actor.get_value(obs, rhs, done).get("value_net1")
            new_val.append(val)
            new_logp.append(logp)

        # Concatenate results
        if self.actor.is_recurrent:
            new_val = [p.view(l, num_proc // self.algo.num_mini_batch, -1) for p in new_val]
            self.data[prl.VAL][:-1] = torch.cat(new_val, dim=1)
            new_logp = [p.view(l, num_proc // self.algo.num_mini_batch, -1) for p in new_logp]
            new_logp = torch.cat(new_logp, dim=1)
        else:
            self.data[prl.VAL][:-1] = torch.cat(new_val, dim=0).view(l, num_proc, 1)
            new_logp = torch.cat(new_logp, dim=0).view(l, num_proc, 1)

        return new_logp

    @torch.no_grad()
    def compute_vtrace(self, clip_rho_thres=1.0, clip_c_thres=1.0):
        """
        Computes V-trace target values and advantage predictions and stores them,
        along with the updated action log probabilities, in `storage`.

        Parameters
        ----------
        clip_rho_thres : float
            V-trace rho threshold parameter.
        clip_c_thres : float
            V-trace c threshold parameter.
        """

        l = self.step if self.step != 0 else self.max_size

        new_action_log_probs = self.get_updated_action_log_probs()

        log_rhos = (new_action_log_probs - self.data[prl.LOGP][:l])
        clipped_rhos = torch.clamp(torch.exp(log_rhos), max=clip_rho_thres)
        clipped_cs = torch.clamp(torch.exp(log_rhos), max=clip_c_thres)

        deltas = clipped_rhos * (
            self.data[prl.RET][:-1] + self.algo.gamma * self.data[prl.VAL][1:]
            - self.data[prl.VAL][:-1])

        acc = torch.zeros_like(self.data[prl.VAL][-1])
        result = []
        for i in reversed(range(l)):
            acc = deltas[i] + self.algo.gamma * clipped_cs[i] * acc * (
                1 - self.data[prl.DONE][i + 1])
            result.append(acc)

        result.reverse()
        result.append(torch.zeros_like(self.data[prl.VAL][-1]))
        vs_minus_v_xs = torch.stack(result)

        vs = torch.add(vs_minus_v_xs, self.data[prl.VAL])
        adv = clipped_rhos * (self.data[prl.RET][:-1] + self.algo.gamma *
                              vs[1:] - self.data[prl.VAL][:-1])

        self.data[prl.RET] = vs
        self.data[prl.LOGP] = new_action_log_probs
        self.data[prl.ADV] = (adv - adv.mean()) / (adv.std() + 1e-8)



import numpy as np
import torch
import torch.nn as nn
from pytorchrl.agent.actors.utils import init


class Categorical(nn.Module):
    """
    Categorical probability distribution.

    Parameters
    ----------
    num_inputs : int
        Size of input feature maps.
    num_outputs : int
        Number of options in output space.
    """
    def __init__(self, num_inputs, num_outputs):
        super(Categorical, self).__init__()

        init_ = lambda m: init(
            m,
            nn.init.orthogonal_,
            lambda x: nn.init.constant_(x, 0),
            gain=np.sqrt(0.01))

        self.linear = init_(nn.Linear(num_inputs, num_outputs))

    def forward(self, x, deterministic=False):
        """
        Predict distribution parameters from x (obs features) and return
        predictions (sampled and clipped), sampled log
        probability and distribution entropy.

        Parameters
        ----------
        x : torch.tensor
            Feature maps extracted from environment observations.
        deterministic : bool
            Whether to randomly sample from predicted distribution or take the mode.

        Returns
        -------
        pred: torch.tensor
            Predicted value.
        clipped_pred: torch.tensor
            Predicted value (clipped to be within [-1, 1] range).
        logp : torch.tensor
            Log probability of `pred` according to the predicted distribution.
        entropy_dist : torch.tensor
            Entropy of the predicted distribution.
        dist : torch.Distribution
            Action probability distribution.
        """

        # Predict distribution parameters
        x = self.linear(x)

        # Create distribution and sample
        dist = torch.distributions.Categorical(logits=x)

        if deterministic:
            pred = clipped_pred = dist.probs.argmax(dim=-1, keepdim=True)
        else:
            pred = clipped_pred = dist.sample().unsqueeze(-1)

        # Action log probability
        # logp = dist.log_prob(pred.squeeze( -1)).unsqueeze(-1)
        logp = dist.log_prob(pred.squeeze(-1)).view(pred.size(0), -1).sum(-1).unsqueeze(-1)

        # Distribution entropy
        entropy_dist = dist.entropy().mean()

        return pred, clipped_pred, logp, entropy_dist, dist

    def evaluate_pred(self, x, pred):
        """
        Return log prob of `pred` under the distribution generated from
        x (obs features). Also return entropy of the generated distribution.

        Parameters
        ----------
        x : torch.tensor
            obs feature map obtained from a policy_net.
        pred : torch.tensor
            Prediction to evaluate.

        Returns
        -------
        logp : torch.tensor
            Log probability of `pred` according to the predicted distribution.
        entropy_dist : torch.tensor
            Entropy of the predicted distribution.
        dist : torch.Distribution
            Action probability distribution.
        """

        # Predict distribution parameters
        x = self.linear(x)

        # Create distribution
        dist = torch.distributions.Categorical(logits=x)

        # Evaluate log prob of under dist
        logp = dist.log_prob(pred.squeeze(-1)).unsqueeze(-1).sum(-1, keepdim=True)

        # Distribution entropy
        entropy_dist = dist.entropy().mean()

        return logp, entropy_dist, dist



"""Utils to map the REINVENT pre-trained network checkpoint to the pytorchrl network."""

import torch
from pytorchrl.envs.generative_chemistry.vocabulary import ReinventVocabulary, LibinventVocabulary

reinvent_weights_mapping = {
    "_embedding.weight": "policy_net.feature_extractor._embedding.weight",
    "_rnn.weight_ih_l0": "policy_net.memory_net._rnn.weight_ih_l0",
    "_rnn.weight_hh_l0": "policy_net.memory_net._rnn.weight_hh_l0",
    "_rnn.bias_ih_l0": "policy_net.memory_net._rnn.bias_ih_l0",
    "_rnn.bias_hh_l0": "policy_net.memory_net._rnn.bias_hh_l0",
    "_rnn.weight_ih_l1": "policy_net.memory_net._rnn.weight_ih_l1",
    "_rnn.weight_hh_l1": "policy_net.memory_net._rnn.weight_hh_l1",
    "_rnn.bias_ih_l1": "policy_net.memory_net._rnn.bias_ih_l1",
    "_rnn.bias_hh_l1": "policy_net.memory_net._rnn.bias_hh_l1",
    "_rnn.weight_ih_l2": "policy_net.memory_net._rnn.weight_ih_l2",
    "_rnn.weight_hh_l2": "policy_net.memory_net._rnn.weight_hh_l2",
    "_rnn.bias_ih_l2": "policy_net.memory_net._rnn.bias_ih_l2",
    "_rnn.bias_hh_l2": "policy_net.memory_net._rnn.bias_hh_l2",
    "_linear.weight": "policy_net.dist.linear.weight",
    "_linear.bias": "policy_net.dist.linear.bias",
}

libinvent_weights_mapping = {
    "_encoder._embedding.0.weight": "policy_net.memory_net._encoder._embedding.0.weight",
    "_encoder._rnn.weight_ih_l0": "policy_net.memory_net._encoder._rnn.weight_ih_l0",
    "_encoder._rnn.weight_hh_l0": "policy_net.memory_net._encoder._rnn.weight_hh_l0",
    "_encoder._rnn.bias_ih_l0": "policy_net.memory_net._encoder._rnn.bias_ih_l0",
    "_encoder._rnn.bias_hh_l0": "policy_net.memory_net._encoder._rnn.bias_hh_l0",
    "_encoder._rnn.weight_ih_l0_reverse": "policy_net.memory_net._encoder._rnn.weight_ih_l0_reverse",
    "_encoder._rnn.weight_hh_l0_reverse": "policy_net.memory_net._encoder._rnn.weight_hh_l0_reverse",
    "_encoder._rnn.bias_ih_l0_reverse": "policy_net.memory_net._encoder._rnn.bias_ih_l0_reverse",
    "_encoder._rnn.bias_hh_l0_reverse": "policy_net.memory_net._encoder._rnn.bias_hh_l0_reverse",
    "_encoder._rnn.weight_ih_l1": "policy_net.memory_net._encoder._rnn.weight_ih_l1",
    "_encoder._rnn.weight_hh_l1": "policy_net.memory_net._encoder._rnn.weight_hh_l1",
    "_encoder._rnn.bias_ih_l1": "policy_net.memory_net._encoder._rnn.bias_ih_l1",
    "_encoder._rnn.bias_hh_l1": "policy_net.memory_net._encoder._rnn.bias_hh_l1",
    "_encoder._rnn.weight_ih_l1_reverse": "policy_net.memory_net._encoder._rnn.weight_ih_l1_reverse",
    "_encoder._rnn.weight_hh_l1_reverse": "policy_net.memory_net._encoder._rnn.weight_hh_l1_reverse",
    "_encoder._rnn.bias_ih_l1_reverse": "policy_net.memory_net._encoder._rnn.bias_ih_l1_reverse",
    "_encoder._rnn.bias_hh_l1_reverse": "policy_net.memory_net._encoder._rnn.bias_hh_l1_reverse",
    "_encoder._rnn.weight_ih_l2": "policy_net.memory_net._encoder._rnn.weight_ih_l2",
    "_encoder._rnn.weight_hh_l2": "policy_net.memory_net._encoder._rnn.weight_hh_l2",
    "_encoder._rnn.bias_ih_l2": "policy_net.memory_net._encoder._rnn.bias_ih_l2",
    "_encoder._rnn.bias_hh_l2": "policy_net.memory_net._encoder._rnn.bias_hh_l2",
    "_encoder._rnn.weight_ih_l2_reverse": "policy_net.memory_net._encoder._rnn.weight_ih_l2_reverse",
    "_encoder._rnn.weight_hh_l2_reverse": "policy_net.memory_net._encoder._rnn.weight_hh_l2_reverse",
    "_encoder._rnn.bias_ih_l2_reverse": "policy_net.memory_net._encoder._rnn.bias_ih_l2_reverse",
    "_encoder._rnn.bias_hh_l2_reverse": "policy_net.memory_net._encoder._rnn.bias_hh_l2_reverse",
    "_decoder._embedding.0.weight": "policy_net.memory_net._decoder._embedding.0.weight",
    "_decoder._rnn.weight_ih_l0": "policy_net.memory_net._decoder._rnn.weight_ih_l0",
    "_decoder._rnn.weight_hh_l0": "policy_net.memory_net._decoder._rnn.weight_hh_l0",
    "_decoder._rnn.bias_ih_l0": "policy_net.memory_net._decoder._rnn.bias_ih_l0",
    "_decoder._rnn.bias_hh_l0": "policy_net.memory_net._decoder._rnn.bias_hh_l0",
    "_decoder._rnn.weight_ih_l1": "policy_net.memory_net._decoder._rnn.weight_ih_l1",
    "_decoder._rnn.weight_hh_l1": "policy_net.memory_net._decoder._rnn.weight_hh_l1",
    "_decoder._rnn.bias_ih_l1": "policy_net.memory_net._decoder._rnn.bias_ih_l1",
    "_decoder._rnn.bias_hh_l1": "policy_net.memory_net._decoder._rnn.bias_hh_l1",
    "_decoder._rnn.weight_ih_l2": "policy_net.memory_net._decoder._rnn.weight_ih_l2",
    "_decoder._rnn.weight_hh_l2": "policy_net.memory_net._decoder._rnn.weight_hh_l2",
    "_decoder._rnn.bias_ih_l2": "policy_net.memory_net._decoder._rnn.bias_ih_l2",
    "_decoder._rnn.bias_hh_l2": "policy_net.memory_net._decoder._rnn.bias_hh_l2",
    "_decoder._attention._attention_linear.0.weight": "policy_net.memory_net._decoder._attention._attention_linear.0.weight",
    "_decoder._attention._attention_linear.0.bias": "policy_net.memory_net._decoder._attention._attention_linear.0.bias",
    "_decoder._linear.weight": "policy_net.dist.linear.weight",
    "_decoder._linear.bias": "policy_net.dist.linear.bias",
}


def adapt_reinvent_checkpoint(file_path):
    """Loads a Reinvent pretrained model and make the necessary changes for it to be compatible with pytorchrl"""

    if torch.cuda.is_available():
        save_dict = torch.load(file_path)
    else:
        save_dict = torch.load(file_path, map_location=lambda storage, loc: storage)

    # Change network weight names
    new_save_dict = {}
    for k in save_dict["network"].keys():
        new_save_dict[reinvent_weights_mapping[k]] = save_dict["network"][k]

    # Temporarily save network weight to /tmp/network_params
    torch.save(new_save_dict, "/tmp/network_params.tmp")

    # Remove unnecessary network parameters
    network_params = save_dict["network_params"]
    network_params.pop("cell_type", None)
    network_params.pop("embedding_layer_size", None)

    return ReinventVocabulary(save_dict["vocabulary"], save_dict["tokenizer"]), \
           save_dict["max_sequence_length"], save_dict["network_params"], "/tmp/network_params.tmp"


def adapt_libinvent_checkpoint(file_path):
    """Loads a Libinvent pretrained model and make the necessary changes for it to be compatible with pytorchrl"""

    if torch.cuda.is_available():
        save_dict = torch.load(file_path)
    else:
        save_dict = torch.load(file_path, map_location=lambda storage, loc: storage)

    # Change network weight names
    new_save_dict = {}
    for k in save_dict["decorator"]["state"].keys():
        new_save_dict[libinvent_weights_mapping[k]] = save_dict["decorator"]["state"][k]

    # Temporarily save network weight to /tmp/network_params
    torch.save(new_save_dict, "/tmp/network_params.tmp")

    vocabulary = LibinventVocabulary(
        save_dict["model"]["vocabulary"].scaffold_vocabulary,
        save_dict["model"]["vocabulary"].scaffold_tokenizer,
        save_dict["model"]["vocabulary"].decoration_vocabulary,
        save_dict["model"]["vocabulary"].decoration_tokenizer,
    )

    return vocabulary, save_dict["model"]["max_sequence_length"], \
           save_dict["decorator"]["params"], "/tmp/network_params.tmp"



import gym
import torch

from pytorchrl.agent.actors.base import Actor
from pytorchrl.agent.actors.world_models import WorldModel
from pytorchrl.agent.actors.world_models.utils import StandardScaler


class ModelBasedPlannerActor(Actor):
    """ Actor Planner class for MB agents. """

    def __init__(self,
                 device,
                 horizon,
                 n_planner,
                 input_space,
                 action_space,
                 algorithm_name,
                 checkpoint=None,
                 world_model_class=None,
                 world_model_kwargs={}):

        super(ModelBasedPlannerActor, self).__init__(
            device=device,
            checkpoint=checkpoint,
            input_space=input_space,
            action_space=action_space)

        if type(action_space) == gym.spaces.discrete.Discrete:
            self.action_dims = action_space.n
            self.action_type = "discrete"
            self.action_low = None
            self.action_high = None
        elif type(action_space) == gym.spaces.box.Box:
            self.action_dims = action_space.shape[0]
            self.action_type = "continuous"
            self.action_low = action_space.low
            self.action_high = action_space.high
        else:
            raise ValueError("Unknown action space")

        self.horizon = horizon
        self.n_planner = n_planner
        self.algorithm_name = algorithm_name

        # ----- World Model ---------------------------------------------------

        self.create_world_dynamics_model(world_model_class, world_model_kwargs)

    @classmethod
    def create_factory(
            cls,
            input_space,
            action_space,
            algorithm_name,
            horizon,
            n_planner,
            restart_model=None,
            world_model_class=None,
            world_model_kwargs={}
    ):
        """
        Returns a function that creates actor critic instances.

        Parameters
        ----------
        horizon : int
            The horizon of online planning.
        n_planner : int
            Number of parallel planned trajectories.
        input_space : gym.Space
            Environment observation space.
        action_space : gym.Space
            Environment action space.
        algorithm_name : str
            Name of the RL algorithm used for learning.
        restart_model : str
            Path to a previously trained Actor checkpoint to be loaded.
        world_model_class : class
            PyTorch nn.Module to approximate world dynamics.
        world_model_kwargs
            Keyword arguments for the world model class.

        Returns
        -------
        create_actor_instance : func
            creates a new OffPolicyActor class instance.
        """

        def create_actor_instance(device):
            """Create and return an actor critic instance."""
            policy = cls(device=device,
                         horizon=horizon,
                         n_planner=n_planner,
                         input_space=input_space,
                         action_space=action_space,
                         algorithm_name=algorithm_name,
                         checkpoint=restart_model,
                         world_model_class=world_model_class,
                         world_model_kwargs=world_model_kwargs)
            policy.to(device)

            try:
                policy.try_load_from_checkpoint()
            except RuntimeError:
                pass

            return policy

        return create_actor_instance

    @property
    def is_recurrent(self):
        """Returns True if the actor network are recurrent."""
        return False

    @property
    def recurrent_hidden_state_size(self):
        """Size of policy recurrent hidden state"""
        return 1

    def actor_initial_states(self, obs):
        """
        Returns all actor inputs required to predict initial action.

        Parameters
        ----------
        obs : torch.tensor
            Initial environment observation.

        Returns
        -------
        obs : torch.tensor
            Initial environment observation.
        rhs : dict
            Initial recurrent hidden state (will contain zeroes).
        done : torch.tensor
            Initial done tensor, indicating the environment is not done.
        """

        if isinstance(obs, dict):
            num_proc = list(obs.values())[0].size(0)
            dev = list(obs.values())[0].device
        else:
            num_proc = obs.size(0)
            dev = obs.device

        done = torch.zeros(num_proc, 1).to(dev)

        try:
            rhs = self.policy_net.memory_net.get_initial_recurrent_state(num_proc).to(dev)
        except Exception:
            rhs = torch.zeros(num_proc, self.recurrent_hidden_state_size).to(dev)

        rhs = {"world_model_rhs": rhs}
        return obs, rhs, done

    def get_prediction(self, obs, act, rhs, done, deterministic=False):
        """
        Predict and return next action, along with other information.

        Parameters
        ----------
        obs : torch.tensor
            Current environment observation.
        act : torch.tensor
            Action to take given obs.
        rhs : dict
            Current recurrent hidden states.
        done : torch.tensor
            Current done tensor, indicating if episode has finished.
        deterministic : bool
            Whether to randomly sample action from predicted distribution or take the mode.

        Returns
        -------
        next_states : torch.Tensor
            Next states.
        rewards : torch.Tensor
            Reward prediction.
        """

        next_states, rewards = self.dynamics_model.predict(obs, act)

        return next_states, rewards

    def create_world_dynamics_model(self, world_model_class, world_model_kwargs):
        """
        Create a world model instance and define it as class attribute under the name `world_wodel`.

        Parameters
        ----------
        world_model_class : class
            WorldModel class
        world_model_kwargs : dict
            WorldModel class arguments
        """

        if world_model_class is None:
            world_model_class = WorldModel

        data_scaler = StandardScaler(self.device)
        dynamics_model = world_model_class(
            self.device, self.input_space, self.action_space,
            data_scaler, **world_model_kwargs).to(self.device)

        setattr(self, "dynamics_model", dynamics_model)



import itertools
import torch


def get_gradients(*nets, grads_to_cpu=False):
    """Gets gradients for all parameters in nets."""

    params = itertools.chain(*[net.parameters() for net in nets])

    grads = []
    for p in params:
        if grads_to_cpu:
            if p.grad is not None:
                grads.append(p.grad.data.cpu().numpy())
            else:
                grads.append(None)
        else:
            if p.grad is not None:
                grads.append(p.grad)

    return grads


def set_gradients(*nets, gradients, device):
    """Sets gradients as the gradient vaues for all parameters in nets."""

    params = itertools.chain(*[net.parameters() for net in nets])

    for g, p in zip(gradients, params):
        if g is not None:
            p.grad = torch.from_numpy(g).to(device)


###### KL DIVERGENCE ###################################################################################################


def bt(m):
    return m.transpose(dim0=-2, dim1=-1)


def btr(m):
    return m.diagonal(dim1=-2, dim2=-1).sum(-1)


def gaussian_kl(mu1, mu2, cov1, cov2):
    """
    Decoupled KL between two multivariate gaussian distribution

    C_μ = KL(f(x|μi,Σi)||f(x|μ,Σi))
    C_Σ = KL(f(x|μi,Σi)||f(x|μi,Σ))

    Adapted from https://github.com/daisatojp/mpo/blob/master/mpo/mpo.py

    Parameters
    ----------
    mu1: torch.tensor
        Mean distribution 1 - (B, n).
    mu2: torch.tensor
        Mean distribution 2 - (B, n).
    cov1: torch.tensor
        Covariance matrix distribution 1 - (B, n, n).
    cov2:
        Covariance matrix distribution 2 - (B, n, n)

    Returns
    -------
    kl_mu: scalar
        Mean term of the KL.
    kl_sigma: scalar
        Covariance term of the KL.

    ref : https://stanford.edu/~jduchi/projects/general_notes.pdf page.13
    """

    n = cov2.size(-1)
    mu1 = mu1.unsqueeze(-1)  # (B, n, 1)
    mu2 = mu2.unsqueeze(-1)  # (B, n, 1)

    sigma1 = cov1 @ bt(cov1)  # (B, n, n)
    sigma2 = cov2 @ bt(cov2)  # (B, n, n)
    sigma1_det = sigma1.det()  # (B,)
    sigma2_det = sigma2.det()  # (B,)
    sigma1_inv = sigma1.inverse()  # (B, n, n)
    sigma2_inv = sigma2.inverse()  # (B, n, n)

    # determinant can be minus due to numerical calculation error
    # https://github.com/daisatojp/mpo/issues/11
    sigma1_det = torch.clamp_min(sigma1_det, 1e-6)
    sigma2_det = torch.clamp_min(sigma2_det, 1e-6)

    inner_mu = ((mu2 - mu1).transpose(-2, -1) @ sigma1_inv @ (mu2 - mu1)).squeeze()  # (B,)
    inner_sigma = torch.log(sigma1_det / sigma2_det) - n + btr(sigma2_inv @ sigma1_inv)  # (B,)
    kl_mu = 0.5 * torch.mean(inner_mu)
    kl_sigma = 0.5 * torch.mean(inner_sigma)

    return kl_mu, kl_sigma



import numpy as np
from pytorchrl.agent.env.openai_baselines_dependencies.vec_envs.vec_env_base import VecEnvBase
from pytorchrl.agent.env.openai_baselines_dependencies.vec_envs.util import copy_obs_dict, dict_to_obs, obs_space_info


class DummyVecEnv(VecEnvBase):
    """
    VecEnv that does runs multiple environments sequentially, that is,
    the step and reset commands are send to one environment at a time.
    Useful when debugging and when num_env == 1 (in the latter case,
    avoids communication overhead)
    """
    def __init__(self, env_fns):
        """
        Arguments:
        env_fns: iterable of callables      functions that build environments
        """
        self.envs = [fn() for fn in env_fns]
        env = self.envs[0]
        VecEnvBase.__init__(self, len(env_fns), env.observation_space, env.action_space)
        obs_space = env.observation_space
        self.keys, shapes, dtypes = obs_space_info(obs_space)
        self.buf_obs = {k: np.zeros((self.num_envs,) + tuple(shapes[k]), dtype=dtypes[k]) for k in self.keys}
        self.buf_dones = np.zeros((self.num_envs,), dtype=np.bool)
        self.buf_rews = np.zeros((self.num_envs,), dtype=np.float32)
        self.buf_infos = [{} for _ in range(self.num_envs)]
        self.actions = None
        self.spec = self.envs[0].spec

    def step_async(self, actions):
        listify = True
        try:
            if len(actions) == self.num_envs:
                listify = False
        except TypeError:
            pass

        if not listify:
            self.actions = actions
        else:
            assert self.num_envs == 1, "actions {} is either not a list or has a wrong size - cannot match to {} environments".format(actions, self.num_envs)
            self.actions = [actions]

    def step_wait(self):
        for e in range(self.num_envs):
            action = self.actions[e]
            # if isinstance(self.envs[e].action_space, spaces.Discrete):
            #    action = int(action)

            obs, self.buf_rews[e], self.buf_dones[e], self.buf_infos[e] = self.envs[e].step(action)
            if self.buf_dones[e]:
                obs = self.envs[e].reset()
            self._save_obs(e, obs)
        return (self._obs_from_buf(), np.copy(self.buf_rews), np.copy(self.buf_dones),
                self.buf_infos.copy())

    def reset(self):
        for e in range(self.num_envs):
            obs = self.envs[e].reset()
            self._save_obs(e, obs)
        return self._obs_from_buf()

    def reset_single_env(self, num_env):
        for e in range(self.num_envs):
            obs = self.envs[e].reset()
            self._save_obs(e, obs)
        return self._obs_from_buf()

    def _save_obs(self, e, obs):
        for k in self.keys:
            if k is None:
                self.buf_obs[k][e] = obs
            else:
                self.buf_obs[k][e] = obs[k]

    def _obs_from_buf(self):
        return dict_to_obs(copy_obs_dict(self.buf_obs))

    def get_images(self):
        return [env.render(mode='rgb_array') for env in self.envs]

    def render(self, mode='human'):
        if self.num_envs == 1:
            return self.envs[0].render(mode=mode)
        else:
            return super().render(mode=mode)



import sys
import time

from pytorchrl.agent.env import VecEnv
from pytorchrl.learner import Learner
from pytorchrl.scheme import Scheme
from pytorchrl.utils import cleanup_log_dir

# backup checks for possible algos and envs
ENVS_LIST = ["gym", "pybullet", "mujoco", "causalworld", "crafter", "atari"]
ON_POLICY_ALGOS = ["PPO", "PPOD"]
OFF_POLICY_ALGOS = ["DQN", "DDPG", "TD3", "SAC", "MPO"]


class Trainer:
    def __init__(self, config, custom_environment_factory=None):
        cleanup_log_dir(config.log_dir)
        # not sure if this is needed anymore since we store it anyway on wandb
        # save_argparse(config, os.path.join(config.log_dir, "conf.yaml"),[])
        self.config = config
        self.custom_environment_factory = custom_environment_factory

        self.train_envs_factory, self.test_envs_factory, self.action_space, self.obs_space = self.setup_environment()
        self.algo_factory, self.algo_name = get_algorithm(self.config)
        self.actor_factory = self.setup_algorithm()
        self.storage_factory = self.setup_storage()
        self.scheme = self.setup_scheme()
        self.learner = self.setup_learner()

    def train(self, wandb):
        iterations = 0
        start_time = time.time()
        while not self.learner.done():

            self.learner.step()

            if iterations % self.config.log_interval == 0:
                log_data = self.learner.get_metrics(add_episodes_metrics=True)
                log_data = {k.split("/")[-1]: v for k, v in log_data.items()}
                wandb.log(log_data, step=self.learner.num_samples_collected)
                self.learner.print_info()

            if iterations % self.config.save_interval == 0:
                save_name = self.learner.save_model()

            if self.config.max_time != -1 and (time.time() - start_time) > self.config.max_time:
                break

            iterations += 1

        print("Finished!")
        sys.exit()

    def setup_environment(self, ):
        if self.custom_environment_factory is not None:
            environment_train_factory = self.custom_environment_factory
        else:
            environment_train_factory, environment_test_factory = self.get_enviroment_factory(
                self.config.environment_name)
        # 1. Define Train Vector of Envs
        train_envs_factory, action_space, obs_space = VecEnv.create_factory(
            env_fn=environment_train_factory,
            env_kwargs=self.config.environment.train_env_config,
            vec_env_size=self.config.num_env_processes, log_dir=self.config.log_dir)

        # 2. Define Test Vector of Envs (Optional)
        if environment_test_factory is not None:
            test_envs_factory, _, _ = VecEnv.create_factory(
                env_fn=environment_test_factory,
                env_kwargs=self.config.environment.test_env_config,
                vec_env_size=self.config.num_env_processes, log_dir=self.config.log_dir)
        else:
            test_envs_factory = lambda v, x, y, z: None
        return train_envs_factory, test_envs_factory, action_space, obs_space

    def setup_algorithm(self, ):
        if self.config.agent.name in ON_POLICY_ALGOS:
            from pytorchrl.agent.actors import OnPolicyActor, get_feature_extractor, get_memory_network
            actor_factory = OnPolicyActor.create_factory(input_space=self.obs_space,
                                                         action_space=self.action_space,
                                                         algorithm_name=self.algo_name,
                                                         restart_model=self.config.restart_model,
                                                         recurrent_net=get_memory_network(self.config.agent.actor.recurrent_net),
                                                         recurrent_net_kwargs=self.config.agent.actor.recurrent_net_kwargs,
                                                         feature_extractor_kwargs=self.config.agent.actor.feature_extractor_kwargs,
                                                         feature_extractor_network=get_feature_extractor(self.config.agent.actor.feature_extractor_network),
                                                         shared_policy_value_network=self.config.agent.actor.shared_policy_value_network)
            return actor_factory
        elif self.config.agent.name in OFF_POLICY_ALGOS:
            from pytorchrl.agent.actors import OffPolicyActor, get_feature_extractor, get_memory_network
            actor_factory = OffPolicyActor.create_factory(input_space=self.obs_space,
                                                          action_space=self.action_space,
                                                          algorithm_name=self.algo_name,
                                                          noise=self.config.agent.noise,
                                                          restart_model=self.config.restart_model,
                                                          sequence_overlap=self.config.agent.sequence_overlap,
                                                          recurrent_net_kwargs=self.config.agent.actor.recurrent_net_kwargs,
                                                          recurrent_net=get_memory_network(self.config.agent.actor.recurrent_net),
                                                          obs_feature_extractor=get_feature_extractor(
                                                              self.config.agent.actor.obs_feature_extractor),
                                                          obs_feature_extractor_kwargs=self.config.agent.actor.obs_feature_extractor_kwargs,
                                                          act_feature_extractor=get_feature_extractor(
                                                              self.config.agent.actor.act_feature_extractor),
                                                          common_feature_extractor=get_feature_extractor(
                                                              self.config.agent.actor.common_feature_extractor),
                                                          common_feature_extractor_kwargs=self.config.agent.actor.common_feature_extractor_kwargs,
                                                          num_critics=self.config.agent.actor.num_critics)
            return actor_factory
        else:
            pass

    def setup_storage(self, ):
        if self.config.agent.storage.name == "replay_buffer":
            from pytorchrl.agent.storages import ReplayBuffer
            return ReplayBuffer.create_factory(size=self.config.agent.storage.size)
        elif self.config.agent.storage.name == "her_buffer":
            from pytorchrl.agent.storages import HERBuffer
            return HERBuffer.create_factory(size=self.config.agent.storage.name)
        elif self.config.agent.storage.name == "vanilla_on_policy":
            from pytorchrl.agent.storages import VanillaOnPolicyBuffer
            return VanillaOnPolicyBuffer.create_factory(size=self.config.agent.storage.size)
        elif self.config.agent.storage.name == "gae_buffer":
            from pytorchrl.agent.storages import GAEBuffer
            return GAEBuffer.create_factory(size=self.config.agent.storage.size,
                                            gae_lambda=self.config.agent.storage.gae_lambda)
        elif self.config.agent.storage.name == "vtrace":
            from pytorchrl.agent.storages import VTraceBuffer
            return VTraceBuffer.create_factory(size=self.config.agent.storage.size)
        else:
            pass

    def setup_scheme(self, ):
        params = {"algo_factory": self.algo_factory,
                  "actor_factory": self.actor_factory,
                  "storage_factory": self.storage_factory,
                  "train_envs_factory": self.train_envs_factory,
                  "test_envs_factory": self.test_envs_factory,
                  "num_col_workers": self.config.scheme.num_col_workers,
                  "col_workers_communication": self.config.scheme.com_col_workers,
                  "col_workers_resources": {"num_cpus": self.config.scheme.col_workers_resources.num_cpus,
                                            "num_gpus": self.config.scheme.col_workers_resources.num_gpus},
                  "num_grad_workers": self.config.scheme.num_grad_workers,
                  "grad_workers_communication": self.config.scheme.com_grad_workers,
                  "grad_workers_resources": {"num_cpus": self.config.scheme.grad_workers_resources.num_cpus,
                                             "num_gpus": self.config.scheme.grad_workers_resources.num_gpus},
                  }
        return Scheme(**params)

    def setup_learner(self, ):
        return Learner(self.scheme,
                       target_steps=self.config.num_env_steps, log_dir=self.config.log_dir)

    # might get extracted to additional utils file 
    def get_enviroment_factory(self, task):
        if task == "pybullet":
            from pytorchrl.envs.pybullet import pybullet_train_env_factory, pybullet_test_env_factory
            return pybullet_train_env_factory, pybullet_test_env_factory
        elif task == "mujoco":
            from pytorchrl.envs.mujoco import mujoco_train_env_factory, mujoco_test_env_factory
            return mujoco_train_env_factory, mujoco_test_env_factory
        elif task == "gym":
            # TODO: maybe create extra factory for gym.
            from pytorchrl.envs.pybullet import pybullet_train_env_factory, pybullet_test_env_factory
            return pybullet_train_env_factory, pybullet_test_env_factory
        elif task == "causalworld":
            from pytorchrl.envs.causal_world import causal_world_train_env_factory
            return causal_world_train_env_factory, causal_world_train_env_factory
        elif task == "atari":
            from pytorchrl.envs.atari import atari_train_env_factory, atari_test_env_factory
            return atari_train_env_factory, atari_test_env_factory
        elif task == "crafter":
            from pytorchrl.envs.crafter import crafter_train_env_factory, crafter_test_env_factory
            return crafter_train_env_factory, crafter_test_env_factory
        elif task == "gen_chem":
            # TODO: 
            # from pytorchrl.envs.generative_chemistry.generative_chemistry_env_factory import generative_chemistry_train_env_factory
            # return generative_chemistry_train_env_factory, None
            raise NotImplementedError
        else:
            assert task in ENVS_LIST, "Selected environment is not in list of possible training environments: {} ".format(
                ENVS_LIST)


def get_algorithm(config):
    if config.agent.name == "PPO":
        from pytorchrl.agent.algorithms import PPO
        algo_factory, algo_name = PPO.create_factory(lr=config.agent.ppo_config.lr,
                                                     eps=config.agent.ppo_config.eps,
                                                     gamma=config.agent.ppo_config.gamma,
                                                     num_epochs=config.agent.ppo_config.num_epochs,
                                                     clip_param=config.agent.ppo_config.clip_param,
                                                     num_mini_batch=config.agent.ppo_config.num_mini_batch,
                                                     test_every=config.agent.ppo_config.test_every,
                                                     max_grad_norm=config.agent.ppo_config.max_grad_norm,
                                                     entropy_coef=config.agent.ppo_config.entropy_coeff,
                                                     value_loss_coef=config.agent.ppo_config.value_loss_coef,
                                                     num_test_episodes=config.agent.ppo_config.num_test_episodes,
                                                     use_clipped_value_loss=config.agent.ppo_config.use_clipped_value_loss,
                                                     )

        return algo_factory, algo_name
    elif config.agent.name == "DDPG":
        from pytorchrl.agent.algorithms import DDPG
        algo_factory, algo_name = DDPG.create_factory(lr_pi=config.agent.ddpg_config.lr_pi,
                                                      lr_q=config.agent.ddpg_config.lr_q,
                                                      gamma=config.agent.ddpg_config.gamma,
                                                      max_grad_norm=config.agent.ddpg_config.max_grad_norm,
                                                      polyak=config.agent.ddpg_config.polyak,
                                                      num_updates=config.agent.ddpg_config.num_updates,
                                                      update_every=config.agent.ddpg_config.update_every,
                                                      test_every=config.agent.ddpg_config.test_every,
                                                      num_test_episodes=config.agent.ddpg_config.num_test_episodes,
                                                      start_steps=config.agent.ddpg_config.start_steps,
                                                      mini_batch_size=config.agent.ddpg_config.mini_batch_size,
                                                      target_update_interval=config.agent.ddpg_config.target_update_interval)
        # TODO: check how to set empty list in hydra
        # policy_loss_addons=config.agent.sac_config.policy_loss_addons)
        return algo_factory, algo_name
    elif config.agent.name == "TD3":
        from pytorchrl.agent.algorithms import TD3
        algo_factory, algo_name = TD3.create_factory(lr_pi=config.agent.td3_config.lr_pi,
                                                     lr_q=config.agent.td3_config.lr_q,
                                                     gamma=config.agent.td3_config.gamma,
                                                     max_grad_norm=config.agent.td3_config.max_grad_norm,
                                                     polyak=config.agent.td3_config.polyak,
                                                     num_updates=config.agent.td3_config.num_updates,
                                                     update_every=config.agent.td3_config.update_every,
                                                     test_every=config.agent.td3_config.test_every,
                                                     num_test_episodes=config.agent.td3_config.num_test_episodes,
                                                     start_steps=config.agent.td3_config.start_steps,
                                                     mini_batch_size=config.agent.td3_config.mini_batch_size,
                                                     target_update_interval=config.agent.td3_config.target_update_interval)
        # TODO: check how to set empty list in hydra
        # policy_loss_addons=config.agent.sac_config.policy_loss_addons)
        return algo_factory, algo_name
    elif config.agent.name == "SAC":
        from pytorchrl.agent.algorithms import SAC
        algo_factory, algo_name = SAC.create_factory(lr_pi=config.agent.sac_config.lr_pi,
                                                     lr_q=config.agent.sac_config.lr_q,
                                                     lr_alpha=config.agent.sac_config.lr_alpha,
                                                     gamma=config.agent.sac_config.gamma,
                                                     polyak=config.agent.sac_config.polyak,
                                                     num_updates=config.agent.sac_config.num_updates,
                                                     test_every=config.agent.sac_config.test_every,
                                                     update_every=config.agent.sac_config.update_every,
                                                     start_steps=config.agent.sac_config.start_steps,
                                                     max_grad_norm=config.agent.sac_config.max_grad_norm,
                                                     initial_alpha=config.agent.sac_config.initial_alpha,
                                                     mini_batch_size=config.agent.sac_config.mini_batch_size,
                                                     num_test_episodes=config.agent.sac_config.num_test_episodes,
                                                     target_update_interval=config.agent.sac_config.target_update_interval
                                                     # TODO: check how to set empty list in hydra
                                                     # policy_loss_addons=config.agent.sac_config.policy_loss_addons 
                                                     )
        return algo_factory, algo_name
    elif config.agent.name == "MPO":
        from pytorchrl.agent.algorithms import MPO
        algo_factory, algo_name = MPO.create_factory(lr_pi=config.agent.mpo_config.lr_pi,
                                                     lr_q=config.agent.mpo_config.lr_q,
                                                     gamma=config.agent.mpo_config.gamma,
                                                     polyak=config.agent.mpo_config.polyak,
                                                     num_updates=config.agent.mpo_config.num_updates,
                                                     test_every=config.agent.mpo_config.test_every,
                                                     update_every=config.agent.mpo_config.update_every,
                                                     start_steps=config.agent.mpo_config.start_steps,
                                                     mini_batch_size=config.agent.mpo_config.mini_batch_size,
                                                     num_test_episodes=config.agent.mpo_config.num_test_episodes,
                                                     target_update_interval=config.agent.mpo_config.target_update_interval,
                                                     dual_constraint=config.agent.mpo_config.dual_constraint,
                                                     kl_mean_constraint=config.agent.mpo_config.kl_mean_constraint,
                                                     kl_var_constraint=config.agent.mpo_config.kl_var_constraint,
                                                     kl_constraint=config.agent.mpo_config.kl_constraint,
                                                     alpha_scale=config.agent.mpo_config.alpha_scale,
                                                     alpha_mean_scale=config.agent.mpo_config.alpha_mean_scale,
                                                     alpha_var_scale=config.agent.mpo_config.alpha_var_scale,
                                                     alpha_mean_max=config.agent.mpo_config.alpha_mean_max,
                                                     alpha_var_max=config.agent.mpo_config.alpha_var_max,
                                                     alpha_max=config.agent.mpo_config.alpha_max,
                                                     mstep_iterations=config.agent.mpo_config.mstep_iterations,
                                                     sample_action_num=config.agent.mpo_config.sample_action_num,
                                                     max_grad_norm=config.agent.mpo_config.max_grad_norm,
                                                     # TODO: check how to set empty list in hydra
                                                     # policy_loss_addons=config.agent.sac_config.policy_loss_addons 
                                                     )
        return algo_factory, algo_name
    else:
        pass



import torch
import numpy as np
import pytorchrl as prl
from pytorchrl.agent.storages.off_policy.nstep_buffer import NStepReplayBuffer as B


def dim0_reshape(tensor, size):
    """
    Reshapes tensor so indices are defined like this:

    00, 01, 02, 03, 04, 05, 06, 07, 08, 09, size + 1, ..., self.max_size
    10, 11, 12, 13, 14, 15, 16, 17, 18, 19, size + 1, ..., self.max_size
    20, 21, 22, 23, 24, 25, 26, 27, 28, 29, size + 1, ..., self.max_size

    """
    return np.moveaxis(tensor, [0, 1], [1, 0])[:, 0: size].reshape(-1, *tensor.shape[2:])


class PERBuffer(B):
    """
    Storage class for Off-Policy algorithms using PER (https://arxiv.org/abs/1707.01495).

    This component extends NStepReplayBuffer, enabling to combine PER with
    n step learning. However, default n_step value is 1, which is equivalent
    to not using n_step learning at all.

    Parameters
    ----------
    size : int
        Storage capacity along time axis.
    device : torch.device
        CPU or specific GPU where data tensors will be placed and class
        computations will take place. Should be the same device where the
        actor model is located.
    envs : VecEnv
        Vector of environments instance.
    actor : Actor
        Actor class instance.
    algorithm : Algorithm
        Algorithm class instance.
    n_step : int or float
        Number of future steps used to computed the truncated n-step return value.
    epsilon : float
        PER epsilon parameter.
    alpha : float
        PER alpha parameter.
    beta : float
        PER beta parameter.
    default_error : int or float
        Default TD error value to use for newly added data samples.
    """

    # Data fields to store in buffer and contained in the generated batches
    storage_tensors = prl.OffPolicyDataKeys

    def __init__(self, size, device, actor, algorithm, envs, n_step=1, epsilon=0.0, alpha=0.0, beta=1.0, default_error=1000000):

        super(PERBuffer, self).__init__(
            size=size, device=device, actor=actor,
            algorithm=algorithm, envs=envs, n_step=n_step)

        self.beta = beta
        self.alpha = alpha
        self.epsilon = epsilon
        self.data["priority"] = None
        self.error = default_error

    @classmethod
    def create_factory(cls, size, n_step=1, epsilon=0.0, alpha=0.0, beta=1.0, default_error=1000000):
        """
        Returns a function that creates PERBuffer instances.

        Parameters
        ----------
        size : int
            Storage capacity along time axis.
        n_step : int or float
            Number of future steps used to computed the truncated n-step return value.
        epsilon : float
            PER epsilon parameter.
        alpha : float
            PER alpha parameter.
        beta : float
            PER beta parameter.
        default_error : int or float
            Default TD error value to use for newly added data samples.

        Returns
        -------
        create_buffer_instance : func
            creates a new PERBuffer class instance.
        """

        def create_buffer(device, actor, algorithm, envs):
            """Create and return a PERBuffer instance."""
            return cls(size, device, actor, algorithm, envs, n_step, epsilon, alpha, beta, default_error)

        return create_buffer

    def get_priority(self, error):
        """Takes in the error of one or more examples and returns the proportional priority"""
        return np.power(error + self.epsilon, self.alpha)

    def get_sequence_priority(self, sequence_data, eta=0.9):
        """ Get priority score for a given data sequence. """
        term1 = eta * np.max(sequence_data, axis=0)
        term2 = (1 - eta) * np.mean(sequence_data, axis=0)
        priority = term1 + term2
        return priority

    def insert_transition(self, sample):
        """
        Store new transition sample.

        Parameters
        ----------
        sample : dict
            Data sample (containing all tensors of an environment transition)
        """

        # Data tensors lazy initialization
        if self.size == 0 and prl.OBS not in self.data.keys():
            self.init_tensors(sample)

        # if recurrent, save fixed-length consecutive overlapping sequences
        if self.recurrent_actor and self.step % self.sequence_length == 0 and self.step != 0:
            next_seq_overlap = self.get_data_slice(self.step - self.overlap_length, self.step)
            self.insert_data_slice(next_seq_overlap)

        # Add obs, rhs, done, act and rew to n_step buffer
        self.n_step_buffer[prl.OBS].append(sample[prl.OBS])
        self.n_step_buffer[prl.REW].append(sample[prl.REW])
        self.n_step_buffer[prl.ACT].append(sample[prl.ACT])
        self.n_step_buffer[prl.RHS].append(sample[prl.RHS])
        self.n_step_buffer[prl.DONE].append(sample[prl.DONE])

        if len(self.n_step_buffer[prl.OBS]) == self.n_step:

            # Add obs2, rhs2 and done2 directly
            for k in (prl.OBS2, prl.RHS2, prl.DONE2):

                if not self.recurrent_actor and k == prl.RHS2:
                    continue

                if isinstance(sample[k], dict):
                    for x, y in sample[k].items():
                        self.data[k][x][self.step] = y.cpu()
                else:
                    self.data[k][self.step] = sample[k].cpu()

            # Compute done and rew
            (self.data[prl.REW][self.step],
             self.data[prl.DONE][self.step]) = self._nstep_return()

            # Get obs, rhs and act from step buffer
            for k in (prl.OBS, prl.RHS, prl.ACT):

                if not self.recurrent_actor and k == prl.RHS:
                    continue

                tensor = self.n_step_buffer[k].popleft()
                if isinstance(tensor, dict):
                    for x, y in tensor.items():
                        self.data[k][x][self.step] = y.cpu()
                else:
                    self.data[k][self.step] = tensor.cpu()

            # Update
            self.step = (self.step + 1) % self.max_size
            self.size = min(self.size + 1, self.max_size)

    def before_gradients(self):
        """
        Steps required before updating actor policy model.
        """
        num_proc = self.data[prl.DONE].shape[1]

        if self.data["priority"] is None:  # Lazy initialization of priority
            if self.recurrent_actor:
                default_priority = self.get_sequence_priority(self.error * np.ones((self.sequence_length)))
            else:
                default_priority = self.error
            self.data["priority"] = default_priority * np.ones((self.max_size, num_proc, 1))

    def after_gradients(self, batch, info):
        """
        Steps required after updating actor policy model

        Parameters
        ----------
        batch : dict
            Data batch used to compute the gradients.
        info : dict
            Additional relevant info from gradient computation.

        Returns
        -------
        info : dict
            info dict updated with relevant info from Storage.
        """

        if "per_weights" in batch.keys() and isinstance(batch["per_weights"], torch.Tensor):

            assert "errors" in info[prl.ALGORITHM].keys(), "TD errors missing!"

            if self.recurrent_actor:

                endpos = int(self.size + self.sequence_length)

                # Get data indices and td errors
                idxs = np.array(batch.pop("idxs")).reshape(-1, self.sequence_length)
                errors = info[prl.ALGORITHM]["errors"].reshape(-1, self.non_overlap_length)

                # Since sequences overlap, update both current sequence and
                # start of the next overlapping sequence
                idxs = np.concatenate([idxs[:, -self.non_overlap_length:], idxs[:, -self.overlap_length:] + self.overlap_length], axis=1)
                errors = torch.cat([errors, errors[:, -self.overlap_length:]], dim=1)

                for i, e in zip(idxs, errors):  # each sequence in the batch

                    # Assign priorities to both end of current sequence
                    # and start of next sequence
                    dim0_reshape(self.data["priority"], endpos)[i] = \
                        self.get_priority(e.unsqueeze(1))

                    # Update current sequence average priority
                    sequence = dim0_reshape(self.data["priority"], endpos)[
                        i - self.overlap_length]
                    pri = self.get_sequence_priority(sequence)
                    dim0_reshape(self.data["priority"], endpos)[
                        i - self.overlap_length] = pri * np.ones(sequence.shape)

                    # Update next sequence average if some overlap
                    if self.overlap_length > 0:
                        sequence = dim0_reshape(
                            self.data["priority"], endpos)[i + self.non_overlap_length]
                        pri = self.get_sequence_priority(sequence)
                        dim0_reshape(self.data["priority"], endpos)[
                            i + self.non_overlap_length] = pri * np.ones(sequence.shape)
            else:
                self.data["priority"][0:self.size].reshape(-1, *self.data["priority"].shape[2:])[
                    batch.pop("idxs")] = self.get_priority(info[prl.ALGORITHM]["errors"])

        return info

    def generate_batches(self, num_mini_batch, mini_batch_size, num_epochs=1):
        """
        Returns a batch iterator to update actor.

        Parameters
        ----------
        num_mini_batch : int
           Number mini batches per epoch.
        mini_batch_size : int
            Number of samples contained in each mini batch.
        num_epochs : int
            Number of epochs.

        Yields
        ------
        batch : dict
            Generated data batches.
        """

        num_proc = self.data[prl.DONE].shape[1]

        for _ in range(num_mini_batch):

            if self.recurrent_actor:  # Batches to feed a recurrent actor

                sequences_x_batch = mini_batch_size // self.sequence_length + 1
                sequences_x_proc = int(self.size / self.sequence_length)

                assert self.size % self.sequence_length == 0, \
                    "Buffer does not contain an integer number of complete rollout sequences"

                # Define batch structure
                batch = {k: {} if not isinstance(self.data[k], dict) else
                    {x: {} for x in self.data[k]} for k in self.data.keys()}

                # Select sequences
                if self.alpha == 0.0:
                    seq_idxs = np.random.randint(0, num_proc * sequences_x_proc, size=sequences_x_batch)
                    per_weigths = 1.0
                else:
                    priors = dim0_reshape(self.data["priority"], self.size)
                    probs = priors / priors.sum()
                    per_weigths = np.power(num_proc * self.size * probs, - self.beta)
                    per_weigths = per_weigths / per_weigths.max()

                    # Trick to allow updating priorities of next overlapping
                    # sequences after gradient computation. Insert some "0.0"
                    # values in per_weights for end-of-row + 1 sequences.
                    per_weigths = np.split(per_weigths, num_proc)
                    per_weigths = [np.concatenate([chunk, np.zeros((
                        self.sequence_length, 1))]) for chunk in per_weigths]
                    per_weigths = np.concatenate(per_weigths)

                    # Trick to allow updating priorities of next overlapping
                    # sequences after gradient computation. Insert some "0.0"
                    # values in probs for end-of-row + 1 sequences.
                    probs = probs[self.sequence_length
                        - 1::self.sequence_length] * self.sequence_length
                    probs = np.split(probs, num_proc)
                    probs = [np.concatenate([chunk, np.zeros(
                        (1, 1))]) for chunk in probs]
                    ext_probs = np.concatenate(probs).squeeze(1)
                    ext_probs = ext_probs / ext_probs.sum() # sanity check
                    seq_idxs = np.random.choice(range(len(ext_probs)), size=sequences_x_batch, p=ext_probs)

                # Get data indexes
                idxs = []
                for idx in seq_idxs:
                    idxs += range(idx * self.sequence_length, (idx + 1) * self.sequence_length)

                if not isinstance(per_weigths, float):
                    per_weigths = torch.as_tensor(per_weigths[idxs], dtype=torch.float32).to(self.device)

                # Fill up batch with data
                for k, v in self.data.items():
                    # Only first recurrent state in each sequence needed
                    positions = seq_idxs * self.sequence_length if k in (prl.RHS, prl.RHS2) else idxs
                    if isinstance(v, dict):
                        for x, y in v.items():
                            t = dim0_reshape(y, self.size + self.sequence_length)[positions]
                            batch[k][x] = torch.as_tensor(t, dtype=torch.float32).to(self.device)
                    else:
                        t = dim0_reshape(v, self.size + self.sequence_length)[positions]
                        batch[k] = torch.as_tensor(t, dtype=torch.float32).to(self.device)

                batch.update({"per_weights": per_weigths, "n_step": self.n_step, "idxs": idxs})
                yield batch

            else:
                batch = {k: {} for k in self.data.keys()}

                if self.alpha == 0.0:
                    samples = np.random.randint(0, num_proc * self.size, size=mini_batch_size)
                    per_weigths = 1.0
                else:
                    priors = self.data["priority"][0:self.size].reshape(-1)
                    probs = priors / priors.sum()
                    samples = np.random.choice(range(num_proc * self.size), size=mini_batch_size, p=probs)
                    per_weigths = np.power(num_proc * self.size * probs, -self.beta)
                    per_weigths = torch.as_tensor(per_weigths / per_weigths.max(), dtype=torch.float32).to(self.device)
                    per_weigths = per_weigths.view(-1, 1)[samples]

                for k, v in self.data.items():

                    if k in (prl.RHS, prl.RHS2):
                        size, idxs = 1, np.array([0])
                    else:
                        size, idxs = self.size, samples

                    if isinstance(v, dict):
                        for x, y in v.items():
                            batch[k][x] = torch.as_tensor(y[0:self.size].reshape(
                                -1, *y.shape[2:])[idxs], dtype=torch.float32).to(self.device)
                    else:
                        batch[k] = torch.as_tensor(v[0:self.size].reshape(
                            -1, *v.shape[2:])[idxs], dtype=torch.float32).to(self.device)

                batch.update({"per_weights": per_weigths, "n_step": self.n_step, "idxs": idxs})
                yield batch



from pytorchrl.agent.actors.memory_networks.gru_net import GruNet
from pytorchrl.agent.actors.memory_networks.lstm_net import LstmNet
from pytorchrl.agent.actors.memory_networks.lstm_encoder_decoder_net import LSTMEncoderDecoder


def get_memory_network(name):
    """Returns model class from name."""
    if name is None:
        return None
    elif name == "GRU":
        return GruNet
    elif name == "LSTM":
        return LstmNet
    elif name == "LSTMEncoderDecoder":
        return LSTMEncoderDecoder
    else:
        raise ValueError("Specified memory net model not found!")



import torch
import torch.nn as nn


class Embedding(nn.Module):
    """
    Embedding feature extractor.

    Parameters
    ----------
    vocabulary_size : int
        Number of tokens in the vocabulary.
    output_size : int
        Embedding layer size.
    """
    def __init__(self, input_space, vocabulary_size, output_size=256):
        super(Embedding, self).__init__()
        self._embedding = nn.Embedding(vocabulary_size, output_size)

    def forward(self, inputs):
        """
        Forward pass Neural Network

        Parameters
        ----------
        inputs : torch.tensor
            Input data.

        Returns
        -------
        out : torch.tensor
            Output feature map.
        """
        input_vector = torch.clamp(inputs, 0.0, self._embedding.num_embeddings).long()
        out = self._embedding(input_vector).squeeze(1)
        return out



import itertools
import numpy as np
from copy import deepcopy
from scipy.optimize import minimize

import torch
import torch.nn as nn
import torch.optim as optim
from torch.distributions.kl import kl_divergence

import pytorchrl as prl
from pytorchrl.agent.algorithms.base import Algorithm
from pytorchrl.agent.algorithms.utils import get_gradients, set_gradients, gaussian_kl
from pytorchrl.agent.algorithms.policy_loss_addons import PolicyLossAddOn


class MPO(Algorithm):
    """
    Maximum a Posteriori Policy Optimization algorithm class.

    Algorithm class to execute MPO, from  A Abdolmaleki et al.
    (https://arxiv.org/abs/1806.06920). Algorithms are modules generally
    required by multiple workers, so MPO.algo_factory(...) returns a function
    that can be passed on to workers to instantiate their own MPO module.

    This code has been adapted from https://github.com/daisatojp/mpo.

    Parameters
    ----------
    device : torch.device
        CPU or specific GPU where class computations will take place.
    envs : VecEnv
        Vector of environments instance.
    actor : Actor
        Actor_critic class instance.
    lr_pi : float
        Policy optimizer learning rate.
    lr_q : float
        Q-nets optimizer learning rate.
    gamma : float
        Discount factor parameter.
    polyak : float
        SAC polyak averaging parameter.
    num_updates : int
        Num consecutive actor updates before data collection continues.
    update_every : int
        Regularity of actor updates in number environment steps.
    start_steps : int
        Num of initial random environment steps before learning starts.
    mini_batch_size : int
        Size of actor update batches.
    target_update_interval : float
        regularity of target nets updates with respect to actor Adam updates.
    num_test_episodes : int
        Number of episodes to complete in each test phase.
    test_every : int
        Regularity of test evaluations in actor updates.
    dual_constraint : float
        Hard constraint of the dual formulation in the E-step corresponding to [2] p.4 ε.
    kl_mean_constraint : float
        Hard constraint of the mean in the M-step corresponding to [2] p.6 ε_μ for continuous action space.
    kl_var_constraint : float
        Hard constraint of the covariance in the M-step corresponding to [2] p.6 ε_Σ for continuous action space.
    kl_constraint : float
        Hard constraint in the M-step corresponding to [2] p.6 ε_π for discrete action space.
    alpha_scale: float
        Scaling factor of the lagrangian multiplier in the M-step for dicrete action spaces.
    alpha_max : float
        Higher bound used for clipping the lagrangian lagrangian in discrete action spaces.
    alpha_mean_scale : float
        Mean scaling factor of the lagrangian multiplier in the M-step for continuous action spaces.
    alpha_var_scale : float
        Varience scaling factor of the lagrangian lagrangian in the M-step for continuous action spaces.
    alpha_mean_max : float
        Higher bound used for clipping the lagrangian lagrangian in continuous action spaces.
    alpha_var_max : float
        Higher bound used for clipping the lagrangian lagrangian in continuous action spaces.
    mstep_iterations : int
        The number of iterations of the M-step
    sample_action_num : int
        For continuous action spaces, number of samples used to compute expected Q scores.
    max_grad_norm : float
        Gradient clipping parameter.
    policy_loss_addons : list
        List of PolicyLossAddOn components adding loss terms to the algorithm policy loss.

    Examples
    --------
    >>> create_algo = MPO.create_factory(
            lr_q=1e-4, lr_pi=1e-4, lr_alpha=1e-4, gamma=0.99, polyak=0.995,
            num_updates=50, update_every=50, test_every=5000, start_steps=20000,
            mini_batch_size=64, alpha=1.0, num_test_episodes=0, target_update_interval=1)
    """

    def __init__(self,
                 device,
                 envs,
                 actor,
                 lr_q=1e-4,
                 lr_pi=1e-4,
                 gamma=0.99,
                 polyak=1.0,
                 num_updates=1,
                 update_every=50,
                 test_every=1000,
                 start_steps=20000,
                 mini_batch_size=64,
                 num_test_episodes=5,
                 target_update_interval=1,
                 dual_constraint=0.1,
                 kl_mean_constraint=0.01,
                 kl_var_constraint=0.0001,
                 kl_constraint=0.01,
                 alpha_scale=10.0,
                 alpha_mean_scale=1.0,
                 alpha_var_scale=100.0,
                 alpha_mean_max=0.1,
                 alpha_var_max=10.0,
                 alpha_max=1.0,
                 mstep_iterations=5,
                 sample_action_num=64,
                 max_grad_norm=0.1,
                 policy_loss_addons=[]):

        # ---- General algo attributes ----------------------------------------

        # Discount factor
        self._gamma = gamma

        # Number of steps collected with initial random policy
        self._start_steps = int(start_steps)

        # Times data in the buffer is re-used before data collection proceeds
        self._num_epochs = int(1)  # Default to 1 for off-policy algorithms

        # Number of data samples collected between network update stages
        self._update_every = int(update_every)

        # Number mini batches per epoch
        self._num_mini_batch = int(num_updates)

        # Size of update mini batches
        self._mini_batch_size = int(mini_batch_size)

        # Number of network updates between test evaluations
        self._test_every = int(test_every)

        # Number of episodes to complete when testing
        self._num_test_episodes = int(num_test_episodes)

        # ---- MPO-specific attributes ----------------------------------------

        self.iter = 0
        self.envs = envs
        self.actor = actor
        self.device = device
        self.polyak = polyak
        self.dual_constraint = dual_constraint
        self.max_grad_norm = max_grad_norm
        self.mstep_iterations = mstep_iterations
        self.sample_action_num = sample_action_num
        self.target_update_interval = target_update_interval

        # For continuous action space
        self.kl_mean_constraint = kl_mean_constraint
        self.kl_var_constraint = kl_var_constraint
        self.alpha_mean_scale = alpha_mean_scale
        self.alpha_var_scale = alpha_var_scale
        self.alpha_mean_max = alpha_mean_max
        self.alpha_var_max = alpha_var_max

        # For discrete action space
        self.alpha_max = alpha_max
        self.kl_constraint = kl_constraint
        self.alpha_scale = alpha_scale

        # Initialize Lagrange Multiplier
        self.eta = np.random.rand()
        self.alpha_mu = 0.0  # lagrangian multiplier for continuous action space in the M-step
        self.alpha_sigma = 0.0  # lagrangian multiplier for continuous action space in the M-step
        self.alpha = 0.0  # lagrangian multiplier for discrete action space in the M-step

        assert hasattr(self.actor, "q1"), "MPO requires q critic (num_critics=1)"

        # Create target networks
        self.actor_targ = deepcopy(actor)

        # Freeze target networks with respect to optimizers
        for p in self.actor_targ.parameters():
            p.requires_grad = False

        # List of parameters for both Q-networks
        q_params = self.actor.q1.parameters()

        # List of parameters for policy network
        p_params = itertools.chain(self.actor.policy_net.parameters())

        # ----- Policy Loss Addons --------------------------------------------

        # Sanity check, policy_loss_addons is a PolicyLossAddOn instance
        # or a list of PolicyLossAddOn instances
        assert isinstance(policy_loss_addons, (PolicyLossAddOn, list)),\
            "MPO policy_loss_addons parameter should be a  PolicyLossAddOn instance " \
            "or a list of PolicyLossAddOn instances"
        if isinstance(policy_loss_addons, list):
            for addon in policy_loss_addons:
                assert isinstance(addon, PolicyLossAddOn), \
                    "MPO policy_loss_addons parameter should be a  PolicyLossAddOn " \
                    "instance or a list of PolicyLossAddOn instances"
        else:
            policy_loss_addons = [policy_loss_addons]

        self.policy_loss_addons = policy_loss_addons
        for addon in self.policy_loss_addons:
            addon.setup(self.actor, self.device)

        # ----- Optimizers ----------------------------------------------------

        self.norm_loss_q = nn.SmoothL1Loss()
        self.pi_optimizer = optim.Adam(p_params, lr=lr_pi)
        self.q_optimizer = optim.Adam(q_params, lr=lr_q)

    @classmethod
    def create_factory(cls,
                       lr_q=1e-4,
                       lr_pi=1e-4,
                       gamma=0.99,
                       polyak=0.995,
                       num_updates=50,
                       test_every=5000,
                       update_every=50,
                       start_steps=1000,
                       mini_batch_size=64,
                       num_test_episodes=5,
                       target_update_interval=1.0,
                       dual_constraint=0.1,
                       kl_mean_constraint=0.01,
                       kl_var_constraint=0.0001,
                       kl_constraint=0.01,
                       alpha_scale=10.0,
                       alpha_mean_scale=1.0,
                       alpha_var_scale=100.0,
                       alpha_mean_max=0.1,
                       alpha_var_max=10.0,
                       alpha_max=1.0,
                       mstep_iterations=5,
                       sample_action_num=64,
                       max_grad_norm=0.1,
                       policy_loss_addons=[]):
        """
        Returns a function to create new MPO instances.

        Parameters
        ----------
        lr_pi : float
            Policy optimizer learning rate.
        lr_q : float
            Q-nets optimizer learning rate.
        gamma : float
            Discount factor parameter.
        polyak : float
            SAC polyak averaging parameter.
        num_updates : int
            Num consecutive actor updates before data collection continues.
        update_every : int
            Regularity of actor updates in number environment steps.
        start_steps : int
            Num of initial random environment steps before learning starts.
        mini_batch_size : int
            Size of actor update batches.
        target_update_interval : float
            regularity of target nets updates with respect to actor Adam updates.
        num_test_episodes : int
            Number of episodes to complete in each test phase.
        test_every : int
            Regularity of test evaluations in actor updates.
        dual_constraint : float
            Hard constraint of the dual formulation in the E-step corresponding to [2] p.4 ε.
        kl_mean_constraint : float
            Hard constraint of the mean in the M-step corresponding to [2] p.6 ε_μ for continuous action space.
        kl_var_constraint : float
            Hard constraint of the covariance in the M-step corresponding to [2] p.6 ε_Σ for continuous action space.
        kl_constraint : float
            Hard constraint in the M-step corresponding to [2] p.6 ε_π for discrete action space.
        alpha_scale: float
            Scaling factor of the lagrangian multiplier in the M-step for dicrete action spaces.
        alpha_max : float
            Higher bound used for clipping the lagrangian lagrangian in discrete action spaces.
        alpha_mean_scale : float
            Mean scaling factor of the lagrangian multiplier in the M-step for continuous action spaces.
        alpha_var_scale : float
            Varience scaling factor of the lagrangian lagrangian in the M-step for continuous action spaces.
        alpha_mean_max : float
            Higher bound used for clipping the lagrangian lagrangian in continuous action spaces.
        alpha_var_max : float
            Higher bound used for clipping the lagrangian lagrangian in continuous action spaces.
        mstep_iterations : int
            The number of iterations of the M-step
        sample_action_num : int
            For continuous action spaces, number of samples used to compute expected Q scores.
        max_grad_norm : float
            Gradient clipping parameter.
        policy_loss_addons : list
            List of PolicyLossAddOn components adding loss terms to the algorithm policy loss.

        Returns
        -------
        create_algo_instance : func
            Function that creates a new MPO class instance.
        algo_name : str
            Name of the algorithm.
        """

        def create_algo_instance(device, actor, envs):
            return cls(lr_q=lr_q,
                       lr_pi=lr_pi,
                       envs=envs,
                       actor=actor,
                       gamma=gamma,
                       device=device,
                       polyak=polyak,
                       test_every=test_every,
                       start_steps=start_steps,
                       num_updates=num_updates,
                       update_every=update_every,
                       mini_batch_size=mini_batch_size,
                       num_test_episodes=num_test_episodes,
                       target_update_interval=target_update_interval,
                       dual_constraint=dual_constraint,
                       kl_mean_constraint=kl_mean_constraint,
                       kl_var_constraint=kl_var_constraint,
                       kl_constraint=kl_constraint,
                       alpha_scale=alpha_scale,
                       alpha_mean_scale=alpha_mean_scale,
                       alpha_var_scale=alpha_var_scale,
                       alpha_mean_max=alpha_mean_max,
                       alpha_var_max=alpha_var_max,
                       policy_loss_addons=policy_loss_addons)

        return create_algo_instance, prl.MPO

    @property
    def gamma(self):
        """Returns discount factor gamma."""
        return self._gamma

    @property
    def start_steps(self):
        """Returns the number of steps to collect with initial random policy."""
        return self._start_steps

    @property
    def num_epochs(self):
        """
        Returns the number of times the whole buffer is re-used before data
        collection proceeds.
        """
        return self._num_epochs

    @property
    def update_every(self):
        """
        Returns the number of data samples collected between
        network update stages.
        """
        return self._update_every

    @property
    def num_mini_batch(self):
        """
        Returns the number of times the whole buffer is re-used before data
        collection proceeds.
        """
        return self._num_mini_batch

    @property
    def mini_batch_size(self):
        """
        Returns the number of mini batches per epoch.
        """
        return self._mini_batch_size

    @property
    def test_every(self):
        """Number of network updates between test evaluations."""
        return self._test_every

    @property
    def num_test_episodes(self):
        """
        Returns the number of episodes to complete when testing.
        """
        return self._num_test_episodes

    @property
    def discrete_version(self):
        """Returns True if action_space is discrete."""
        return self.actor.action_space.__class__.__name__ == "Discrete"

    def acting_step(self, obs, rhs, done, deterministic=False):
        """
        MPO acting function.

        Parameters
        ----------
        obs : torch.tensor
            Current world observation
        rhs : torch.tensor
            RNN recurrent hidden state (if policy is not a RNN, rhs will contain zeroes).
        done : torch.tensor
            1.0 if current obs is the last one in the episode, else 0.0.
        deterministic : bool
            Whether to randomly sample action from predicted distribution or taking the mode.

        Returns
        -------
        action : torch.tensor
            Predicted next action.
        clipped_action : torch.tensor
            Predicted next action (clipped to be within action space).
        rhs : torch.tensor
            Policy recurrent hidden state (if policy is not a RNN, rhs will contain zeroes).
        other : dict
            Additional MPO predictions, which are not used in other algorithms.
        """

        with torch.no_grad():
            (action, clipped_action, logp_action, rhs,
             entropy_dist, dist) = self.actor.get_action(
                obs, rhs, done, deterministic=deterministic)

        return action, clipped_action, rhs, {}

    def compute_loss_q(self, batch, n_step=1, per_weights=1):
        """
        Calculate MPO Q-nets loss

        Parameters
        ----------
        batch : dict
            Data batch dict containing all required tensors to compute MPO losses.
        n_step : int or float
            Number of future steps used to computed the truncated n-step return value.
        per_weights :
            Prioritized Experience Replay (PER) important sampling weights or 1.0.

        Returns
        -------
        loss_q : torch.tensor
            Q-net loss.
        errors : torch.tensor
            TD errors.
        """

        # [2] 3 Policy Evaluation (Step 1)

        o, rhs, d = batch[prl.OBS], batch[prl.RHS], batch[prl.DONE]
        a, r = batch[prl.ACT], batch[prl.REW]
        o2, rhs2, d2 = batch[prl.OBS2], batch[prl.RHS2], batch[prl.DONE2]

        bs, ds = o.shape[0], o.shape[-1]

        if self.discrete_version:

            # Q-values for all actions
            q1 = self.actor.get_q_scores(o, rhs, d).get("q1")
            q1 = q1.gather(1, a.long())

            # Bellman backup for Q functions
            with torch.no_grad():
                # Target actions come from *current* policy
                a2, _, _, _, _, dist = self.actor_targ.get_action(o2, rhs2, d2)

                N = dist.probs.shape[-1]  # num actions
                actions = torch.arange(N)[..., None].expand(-1, bs).to(self.device)  # (da, bs)
                p_a2 = dist.expand((N, bs)).log_prob(actions).exp().transpose(0, 1)  # (bs, da)

                # Target Q-values
                q1_pi_targ = self.actor_targ.get_q_scores(o2, rhs2, d2).get("q1")
                q_pi_targ = (p_a2 * q1_pi_targ).sum(dim=1, keepdim=True)

                assert r.shape == q_pi_targ.shape
                backup = r + (self.gamma ** n_step) * (1 - d2) * q_pi_targ

        else:

            N = self.sample_action_num
            da = a.shape[-1]  # num action dimensions

            # Q-values for all actions
            q1 = self.actor.get_q_scores(o, rhs, d, a).get("q1")

            # Bellman backup for Q functions
            with torch.no_grad():

                # Target actions come from *current* policy
                a2, _, logp_a2, _, _, dist = self.actor_targ.get_action(o2, rhs2, d2)

                sampled_actions = dist.sample((N,)).transpose(0, 1)  # (bs, N, da)
                expanded_obs2 = o2[:, None, :].expand(-1, N, -1)  # (bs, N, ds)
                expanded_d2 = d2[:, None, :].expand(-1, N, -1)  # (bs, N, 1)
                expanded_rhs2 = {k: v[:, None, :].expand(-1, N, -1) for k, v in rhs2.items()}
                expanded_reshaped_rhs2 = {k: v.reshape(-1, v.shape[-1]) for k, v in expanded_rhs2.items()}

                next_q1 = self.actor_targ.get_q_scores(
                    expanded_obs2.reshape(-1, ds),  # (N * bs, ds)
                    expanded_reshaped_rhs2,  # get expanded rhs
                    expanded_d2.reshape(-1, 1),  # (N * bs, ds)
                    sampled_actions.reshape(-1, da),  # (N * bs, da)
                ).get("q1")  # (N * bs, 1)

                expected_next_q1 = next_q1.reshape(bs, N).mean(dim=1, keepdim=True)  # (B,)
                q_pi_targ = expected_next_q1

                backup = r + (self.gamma ** n_step) * (1 - d2) * q_pi_targ

        # MSE loss against Bellman backup
        loss_q = 0.5 * (((q1 - backup) ** 2) * per_weights).mean()
        # loss_q = self.norm_loss_q(backup, q1)

        errors = ((q1 - backup).abs()).detach().cpu()

        return loss_q, errors

    def compute_loss_pi(self, batch, per_weights=1):
        """
        Calculate MPO policy loss.

        Parameters
        ----------
        batch : dict
            Data batch dict containing all required tensors to compute MPO losses.
        per_weights :
            Prioritized Experience Replay (PER) important sampling weights or 1.0.

        Returns
        -------
        loss_policy : torch.tensor
            MPO policy loss.
        """

        o, rhs, d, a = batch[prl.OBS], batch[prl.RHS], batch[prl.DONE], batch[prl.ACT]
        bs, ds = o.shape[0], o.shape[-1]

        # E-Step of Policy Improvement
        # [2] 4.1 Finding action weights (Step 2)

        _, _, _, _, _, dist_targ = self.actor_targ.get_action(o, rhs, d)

        if self.discrete_version:

            N = dist_targ.probs.shape[-1]  # num possible actions actions, env.action_space.n

            # for each state in the batch, any possible action
            actions = torch.arange(N)[..., None].expand(N, bs).to(self.device)  # (N, bs)
            dist_targ_probs = dist_targ.expand((N, bs)).log_prob(actions).exp()  # (N, bs)

            target_q1 = self.actor_targ.get_q_scores(o, rhs, d).get("q1")  # (bs, N)
            target_q1 = target_q1.transpose(1, 0)  # (N, bs)

            b_prob_np = dist_targ_probs.cpu().transpose(0, 1).numpy()  # (bs, N)
            target_q1_np = target_q1.cpu().transpose(0, 1).numpy()  # (bs, N)

        else:

            N = self.sample_action_num
            da = a.shape[-1]  # num action dimensions
            sampled_actions = dist_targ.sample((N,))  # (N, bs, da)
            expanded_obs = o[None, ...].expand(N, -1, -1)  # (N, bs, ds)
            expanded_d = d[None, ...].expand(N, -1, -1)  # (N, bs, 1)
            expanded_rhs = {k: v[None, ...].expand(N, -1, -1) for k, v in rhs.items()}
            expanded_reshaped_rhs = {k: v.reshape(-1, v.shape[-1]) for k, v in expanded_rhs.items()}

            target_q1 = self.actor_targ.get_q_scores(
                expanded_obs.reshape(-1, ds),  # (N * bs, ds)
                expanded_reshaped_rhs,  # get expanded rhs
                expanded_d.reshape(-1, 1),  # (N * bs, ds)
                sampled_actions.reshape(-1, da),  # (N * bs, da)
            ).get("q1")

            target_q1 = target_q1.reshape(N, bs)  # (N, bs)
            target_q1_np = target_q1.cpu().transpose(0, 1).numpy()  # (bs, N)

        # https://arxiv.org/pdf/1812.02256.pdf
        # [2] 4.1 Finding action weights (Step 2)
        #   Using an exponential transformation of the Q-values
        if self.discrete_version:
            def dual(eta):
                """
                dual function of the non-parametric variational
                g(η) = η*ε + η*mean(log(sum(π(a|s)*exp(Q(s, a)/η))))
                We have to multiply π by exp because this is expectation.
                This equation is correspond to last equation of the [2] p.15
                For numerical stabilization, this can be modified to
                Qj = max(Q(s, a), along=a)
                g(η) = η*ε + mean(Qj, along=j) + η*mean(log(sum(π(a|s)*(exp(Q(s, a)-Qj)/η))))
                """
                max_q = np.max(target_q1_np, 1)
                return eta * self.dual_constraint + np.mean(max_q) + eta * np.mean(np.log(np.sum(
                    b_prob_np * np.exp((target_q1_np - max_q[:, None]) / eta), axis=1)))
        else:  # discrete action space
            def dual(eta):
                """
                dual function of the non-parametric variational
                Q = target_q_np  (K, N)
                g(η) = η*ε + η*mean(log(mean(exp(Q(s, a)/η), along=a)), along=s)
                For numerical stabilization, this can be modified to
                Qj = max(Q(s, a), along=a)
                g(η) = η*ε + mean(Qj, along=j) + η*mean(log(mean(exp((Q(s, a)-Qj)/η), along=a)), along=s)
                """
                max_q = np.max(target_q1_np, 1)
                return eta * self.dual_constraint + np.mean(max_q) + eta * np.mean(np.log(
                    np.mean(np.exp((target_q1_np - max_q[:, None]) / eta), axis=1)))

        bounds = [(1e-6, None)]
        res = minimize(dual, np.array([self.eta]), method='SLSQP', bounds=bounds)
        self.eta = res.x[0]
        qij = torch.softmax(target_q1 / self.eta, dim=0)  # (N, bs)

        # M-Step of Policy Improvement
        # [2] 4.2 Fitting an improved policy (Step 3)
        for _ in range(self.mstep_iterations):
            if self.discrete_version:

                _, _, _, _, _, dist = self.actor.get_action(o, rhs, d)
                loss_pi = torch.mean(qij * dist.expand((N, bs)).log_prob(actions))
                kl = kl_divergence(dist, dist_targ).mean()

                # Update lagrange multipliers by gradient descent
                # this equation is derived from last eq of [2] p.5,
                # just differentiate with respect to α
                # and update α so that the equation is to be minimized.
                self.alpha -= self.alpha_scale * (self.kl_constraint - kl).detach().item()
                self.alpha = np.clip(self.alpha, 0.0, self.alpha_max)

                # last eq of [2] p.5
                loss_policy = -(loss_pi + self.alpha * (self.kl_constraint - kl))

            else:

                _, _, _, _, _, dist = self.actor.get_action(o, rhs, d)

                loss_pi = torch.mean(
                    qij * (
                            dist_targ.expand((N, bs, da)).log_prob(sampled_actions).sum(-1)  # (N, K)
                            + dist.expand((N, bs, da)).log_prob(sampled_actions).sum(-1)  # (N, K)
                    )
                )

                # Define diag covariance matrices
                cov1 = torch.eye(dist.variance.shape[-1]).to(self.device) * dist.variance.unsqueeze(
                    2).expand(*dist.variance.size(), dist.variance.size(1))
                cov2 = torch.eye(dist_targ.variance.shape[-1]).to(self.device) * dist_targ.variance.unsqueeze(
                    2).expand(*dist_targ.variance.size(), dist_targ.variance.size(1))
                kl_mu, kl_sigma = gaussian_kl(dist_targ.mean, dist.mean, cov1, cov2)

                if np.isnan(kl_mu.item()):  # This should not happen
                    raise RuntimeError('kl_mu is nan')
                if np.isnan(kl_sigma.item()):  # This should not happen
                    raise RuntimeError('kl_sigma is nan')

                # Update lagrange multipliers by gradient descent
                # this equation is derived from last eq of [2] p.5, just differentiate with
                # respect to α and update α so that the equation is to be minimized.
                self.alpha_mu -= self.alpha_mean_scale * (self.kl_mean_constraint - kl_mu).detach().item()
                self.alpha_sigma -= self.alpha_var_scale * (self.kl_var_constraint - kl_mu).detach().item()

                self.alpha_mu = np.clip(self.alpha_mu, 0.0, self.alpha_mean_max)
                self.alpha_sigma = np.clip(self.alpha_sigma, 0.0, self.alpha_var_max)

                # last eq of [2] p.5
                loss_policy = -(loss_pi + self.alpha_mu * (self.kl_mean_constraint - kl_mu) +
                                self.alpha_sigma * (self.kl_var_constraint - kl_sigma))

        addons_info = {}
        for addon in self.policy_loss_addons:
            addon_loss, addons_info = addon.compute_loss_term(data, dist, addons_info)
            loss_policy += addon_loss

        return loss_policy, addons_info

    def compute_gradients(self, batch, grads_to_cpu=True):
        """
        Compute loss and compute gradients but don't do optimization step,
        return gradients instead.

        Parameters
        ----------
        batch : dict
            data batch containing all required tensors to compute MPO losses.
        grads_to_cpu : bool
            If gradient tensor will be sent to another node, need to be in CPU.

        Returns
        -------
        grads : list of tensors
            List of actor gradients.
        info : dict
            Dict containing current MPO iteration information.
        """

        # Recurrent burn-in
        if self.actor.is_recurrent:
            batch = self.actor.burn_in_recurrent_states(batch)

        # N-step returns
        n_step = batch["n_step"] if "n_step" in batch else 1.0

        # PER
        per_weights = batch["per_weights"] if "per_weights" in batch else 1.0

        # Run one gradient descent step for Q1 and Q2
        loss_q, errors = self.compute_loss_q(batch, n_step, per_weights)
        self.q_optimizer.zero_grad()
        loss_q.backward(retain_graph=True)
        nn.utils.clip_grad_norm_(self.actor.q1.parameters(), self.max_grad_norm)
        q_grads = get_gradients(self.actor.q1, self.actor.q2, grads_to_cpu=grads_to_cpu)

        # Freeze Q-networks so you don't waste computational effort
        # computing gradients for them during the policy learning step.
        for p in self.actor.q1.parameters():
            p.requires_grad = False

        # Run one gradient descent step for pi.
        loss_pi, addons_info = self.compute_loss_pi(batch, per_weights)

        self.pi_optimizer.zero_grad()
        loss_pi.backward()
        nn.utils.clip_grad_norm_(self.actor.policy_net.parameters(), self.max_grad_norm)
        pi_grads = get_gradients(self.actor.policy_net, grads_to_cpu=grads_to_cpu)

        for p in self.actor.q1.parameters():
            p.requires_grad = True

        info = {
            "loss_q": loss_q.detach().item(),
            "loss_pi": loss_pi.detach().item(),
        }

        if "per_weights" in batch:
            info.update({"errors": errors})

        grads = {"q_grads": q_grads, "pi_grads": pi_grads}

        info.update(addons_info)

        return grads, info

    def update_target_networks(self):
        """Update actor critic target networks with polyak averaging."""
        if self.iter % self.target_update_interval == 0:
            with torch.no_grad():
                for p, p_targ in zip(self.actor.parameters(), self.actor_targ.parameters()):
                    p_targ.data.mul_(self.polyak)
                    p_targ.data.add_((1 - self.polyak) * p.data)

    def apply_gradients(self, gradients=None):
        """
        Take an optimization step, previously setting new gradients if provided.
        Update also target networks.

        Parameters
        ----------
        gradients : list of tensors
            List of actor gradients.
        """
        if gradients:
            set_gradients(
                self.actor.policy_net,
                gradients=gradients["pi_grads"], device=self.device)
            set_gradients(
                self.actor.q1, self.actor.q2,
                gradients=gradients["q_grads"], device=self.device)

        self.q_optimizer.step()
        self.pi_optimizer.step()

        # Update target networks by polyak averaging.
        self.iter += 1
        self.update_target_networks()

    def set_weights(self, actor_weights):
        """
        Update actor with the given weights. Update also target networks.

        Parameters
        ----------
        actor_weights : dict of tensors
            Dict containing actor weights to be set.
        """
        self.actor.load_state_dict(actor_weights)

        # Update target networks by polyak averaging.
        self.iter += 1
        self.update_target_networks()

    def update_algorithm_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of the algorithm, change its value
        to `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Worker.algo attribute name
        new_parameter_value : int or float
            New value for `parameter_name`.
        """
        if hasattr(self, parameter_name):
            setattr(self, parameter_name, new_parameter_value)
        if parameter_name == "lr":
            for param_group in self.pi_optimizer.param_groups:
                param_group['lr'] = new_parameter_value
            for param_group in self.q_optimizer.param_groups:
                param_group['lr'] = new_parameter_value



import gym
import numpy as np
import torch.nn as nn
from pytorchrl.agent.actors.utils import init
from pytorchrl.agent.actors.feature_extractors.utils import get_gain


class MLP(nn.Module):
    """
    Multilayer Perceptron network

    Parameters
    ----------
    input_space : gym.Space
        Environment observation space.
    hidden_sizes : list
        Hidden layers sizes.
    output_size : int
        Size of output feature map.
    activation : func
        Non-linear activation function.
    final_activation : bool
        Whether or not to apply activation function after last layer.
    """
    def __init__(self, input_space, hidden_sizes=[256, 256], output_size=256, activation=nn.ReLU, final_activation=True):
        super(MLP, self).__init__()

        if isinstance(input_space, gym.Space):
            input_shape = input_space.shape
        else:
            input_shape = input_space

        # Define feature extractor
        layers = []
        sizes = [np.prod(input_shape)] + hidden_sizes + [output_size]
        for j in range(len(sizes) - 1):
            layers += [nn.Linear(sizes[j], sizes[j + 1])]
            if j < len(sizes) - 2 or final_activation:
                layers += [activation()]
        self.feature_extractor = nn.Sequential(*layers)

        for layer in self.feature_extractor.children():
            if isinstance(layer, nn.Linear):
                nn.init.orthogonal_(layer.weight, gain=get_gain(activation))
                layer.bias.data.zero_()

        self.train()

    def forward(self, inputs):
        """
        Forward pass Neural Network

        Parameters
        ----------
        inputs : torch.tensor
            Input data.

        Returns
        -------
        out : torch.tensor
            Output feature map.
        """
        inputs = inputs.view(inputs.size(0), -1)
        out = self.feature_extractor(inputs)
        return out



import os
from pytorchrl.envs.generative_chemistry.libinvent.rnn_environment import GenChemEnv as RnnLibInvent
from pytorchrl.envs.generative_chemistry.vocabulary import SMILESTokenizer, create_vocabulary


def libinvent_train_env_factory(
        scoring_function, vocabulary, smiles_max_length=200,  scaffolds=[],
        randomize_scaffolds=False, reaction_filters=[], accumulate_obs=False):
    """
    Create train GenChem environment.

    Parameters
    ----------
    scoring_function : func
        Function that given a SMILE, returns a score.
    vocabulary : class
        Class that stores the tokens and allows their conversion to vocabulary indexes.
    smiles_max_length : int
        Maximum length allowed for the generated SMILES. Equivalent to max episode length.
    scaffolds : list
        List of LibInvent scaffolds to be decorated.
    randomize_scaffolds : bool
        Whether or not a random SMILES representation of the scaffolds should be used at each episode.
        Crucially, this is not yet possible if a selective reaction filter is imposed.
    reaction_filters : list
        Reaction filters favored in the reward function.
    accumulate_obs : bool
        If True, obs are accumulated at every time step (e.g. C, CC, CCC, ...)

    Returns
    -------
    env : gym.Env
        Train environment.
    """

    env = RnnLibInvent(
        scoring_function=scoring_function,
        vocabulary=vocabulary,
        max_length=smiles_max_length,
        scaffolds=scaffolds,
        randomize_scaffolds=randomize_scaffolds,
        reactions=reaction_filters,
    )

    return env
